# Copyright contributors to the TSFM project
#

"""Tests the time series imputation preprocessor and functions"""

import numpy as np
import pytest

from tsfm_public import TSPulseConfig, TSPulseForReconstruction
from tsfm_public.toolkit.time_series_imputation_pipeline import TimeSeriesImputationPipeline
from tsfm_public.toolkit.time_series_preprocessor import TimeSeriesPreprocessor


@pytest.fixture(scope="module")
def tspulse_model(etth_data):
    _, _, params = etth_data
    conf = TSPulseConfig(
        context_length=512,
        d_model=24,
        decoder_d_model=24,
        num_patches=128,
        patch_stride=8,
        fft_time_add_forecasting_pt_loss=False,
        num_input_channels=len(params["target_columns"]),
        mask_type="user",
        patch_length=8,
        mask_block_length=2,
        mode="common_channel",
        gated_attn=True,
        use_positional_encoding=False,
        self_attn=False,
        head_aggregation="max_pool",
        fuse_fft=True,
        use_learnable_mask_token=True,
        prediction_length=4,
        channel_mix_init="identity",
    )
    model = TSPulseForReconstruction(conf)
    return model


@pytest.fixture(scope="module")
def etth_missing_data(etth_data):
    train_data, test_data, params = etth_data
    test_data = test_data.copy()

    # quick and dirty random missing
    n = 20
    rng = np.random.default_rng(seed=42)
    inds = rng.integers(test_data.index[0], test_data.index[-1], n)
    sizes = rng.integers(1, 9, n)
    cols = rng.integers(0, len(params["target_columns"]), n)

    for i, s, c in zip(inds, sizes, cols):
        test_data.loc[i : i + s, params["target_columns"][c]] = np.nan

    return train_data, test_data, params


def test_time_series_imputation_pipeline_defaults(tspulse_model, etth_missing_data):
    train_data, test_data, params = etth_missing_data
    test_data = test_data.copy()

    tsp = TimeSeriesPreprocessor(target_columns=params["target_columns"], scaling=True)
    tsp.train(train_data)

    pipe = TimeSeriesImputationPipeline(tspulse_model, feature_extractor=tsp, device="cpu")

    test_imputed = pipe(test_data)

    assert test_imputed.shape[0] == test_data.shape[0]
    assert all(t in test_imputed.columns for t in params["target_columns"])
    assert all(f"{t}_imputed" in test_imputed.columns for t in params["target_columns"])


def test_imputation_pipeline_outputs_for_nan(tspulse_model, etth_missing_data):
    train_data, test_data, params = etth_missing_data
    test_data = test_data.copy()

    tsp = TimeSeriesPreprocessor(target_columns=params["target_columns"], scaling=True)
    tsp.train(train_data)

    pipe = TimeSeriesImputationPipeline(tspulse_model, feature_extractor=tsp, device="cpu")

    test_imputed = pipe(test_data)

    imputed_columns = []
    for c in params["target_columns"]:
        imputed_columns.append(f"{c}_imputed")

    assert not test_imputed[imputed_columns].isnull().values.any()


def test_imputation_pipeline_outputs_for_original_values(tspulse_model, etth_missing_data):
    train_data, test_data, params = etth_missing_data
    test_data = test_data.copy()

    tsp = TimeSeriesPreprocessor(target_columns=params["target_columns"], scaling=True)
    tsp.train(train_data)

    pipe = TimeSeriesImputationPipeline(tspulse_model, feature_extractor=tsp, device="cpu")

    test_imputed = pipe(test_data)

    for col in params["target_columns"]:
        imp_col = f"{col}_imputed"

        non_missing_loc = test_imputed[col].notna()
        assert (test_imputed.loc[non_missing_loc, col] == test_imputed.loc[non_missing_loc, imp_col]).all()


def test_idempotency_on_fully_observed_data(tspulse_model, etth_data):
    train_data, test_data, params = etth_data
    test_data = test_data.copy()

    assert not test_data.isna().any().any()  # no missing values

    tsp = TimeSeriesPreprocessor(target_columns=params["target_columns"], scaling=True)
    tsp.train(train_data)

    pipe = TimeSeriesImputationPipeline(tspulse_model, feature_extractor=tsp, device="cpu")

    test_imputed = pipe(test_data)

    for col in params["target_columns"]:
        imputed_col = f"{col}_imputed"
        assert imputed_col in test_imputed.columns, f"Missing imputed column: {imputed_col}"

        original_vals = test_imputed[col].values
        imputed_vals = test_imputed[imputed_col].values

        assert np.allclose(
            original_vals, imputed_vals
        ), f"Imputed column '{imputed_col}' differs from original column '{col}'"



"""Service handler utils for TSFM models"""

import importlib
import logging
import pathlib
from typing import Any, Dict, Optional, Union

import transformers
from transformers import (
    AutoConfig,
    AutoModel,
    PretrainedConfig,
    PreTrainedModel,
)

import tsfm_public
from tsfm_public.toolkit.hf_util import register_config

from . import TSFM_ALLOW_LOAD_FROM_HF_HUB


LOCAL_FILES_ONLY = not TSFM_ALLOW_LOAD_FROM_HF_HUB


LOGGER = logging.getLogger(__file__)


def load_config(
    model_path: Union[str, pathlib.Path],
    model_type: Optional[str] = None,
    model_config_name: Optional[str] = None,
    module_path: Optional[str] = None,
    **extra_config_kwargs: Dict[str, Any],
) -> PretrainedConfig:
    """Load configuration

    Attempts to load the configuration, if it is not loadable, then we register it with the AutoConfig mechanism.

    Args:
        model_path (pathlib.Path): The path from which to load the config.
        model_type (Optional[str], optional): The type of the model, from the model implementation. Defaults to None.
        model_config_name (Optional[str], optional): The name of configuration class for the model. Defaults to None.
        module_path (Optional[str], optional): Python module path that can be used to load the
            config/model. Defaults to None.

    Returns:
        PretrainedConfig: The configuration object corresponding to the pretrained model.
    """
    # load config first try autoconfig, if not then we register and load
    try:
        conf = AutoConfig.from_pretrained(model_path, local_files_only=LOCAL_FILES_ONLY, **extra_config_kwargs)
    except (KeyError, ValueError) as exc:  # determine error raised by autoconfig
        if model_type is None or model_config_name is None or module_path is None:
            raise ValueError("model_type, model_config_name, and module_path should be specified.") from exc

        register_config(model_type, model_config_name, module_path)
        conf = AutoConfig.from_pretrained(model_path, local_files_only=LOCAL_FILES_ONLY, **extra_config_kwargs)

    return conf


def _get_model_class(config: PretrainedConfig, module_path: Optional[str] = None) -> type:
    """Helper to find model class based on config object

    First the module_path will be checked if it can be loaded in the current environment. If not
    then the transformers library will be used.

    Args:
        config (PretrainedConfig): HF configuration for the model.
        module_path (Optional[str], optional): Python module path that can be used to load the
            config/model. Defaults to None.

    Raises:
        AttributeError: Raised if the module at module_path cannot be loaded.
        AttributeError: If the architecture provided by the config cannot be loaded from
            the module.

    Returns:
        type: The class for the model.
    """
    if module_path is not None:
        try:
            mods = [importlib.import_module(module_path)]
        except ModuleNotFoundError as exc:
            raise AttributeError("Could not load module '{module_path}'.") from exc
    else:
        mods = [transformers, tsfm_public]

    # get architecture from model config
    architectures = getattr(config, "architectures", [])
    for arch in architectures:
        model_class = None
        for mod in mods:
            model_class = getattr(mod, arch, None)
            if model_class is not None:
                return model_class
        raise AttributeError(f"Could not load model class for architecture '{arch}'.")

        # try:
        #     model_class = getattr(mod, arch)
        #     return model_class
        # except AttributeError as exc:
        #     # catch specific error import error or attribute error
        #     raise AttributeError(f"Could not load model class for architecture '{arch}'.") from exc


def load_model(
    model_path: Union[str, pathlib.Path],
    config: Optional[PretrainedConfig] = None,
    module_path: Optional[str] = None,
) -> PreTrainedModel:
    """Load a pretrained model.
    If module_path is provided, load the model using the provided module path.

    Args:
        model_path (Union[str, pathlib.Path]): Path to a location where the model can be loaded.
        config (Optional[PretrainedConfig], optional): HF Configuration object. Defaults to None.
        module_path (Optional[str], optional): Python module path that can be used to load the
            config/model. Defaults to None.

    Raises:
        ValueError: Raised if loading from a module_path and a configuration object is not provided.

    Returns:
        PreTrainedModel: The loaded pretrained model.
    """

    if module_path is not None and config is None:
        return ValueError("Config must be provided when loading from a custom module_path")

    if config is not None:
        model_class = _get_model_class(config, module_path=module_path)
        LOGGER.info(f"Found model class: {model_class.__name__}")
        return model_class.from_pretrained(
            model_path,
            config=config,
            local_files_only=LOCAL_FILES_ONLY,
        )

    return AutoModel.from_pretrained(
        model_path,
        local_files_only=LOCAL_FILES_ONLY,
    )



# Copyright contributors to the TSFM project
#
"""Tsfmfinetuning Runtime"""

import logging
from pathlib import Path
from typing import Tuple, Union

import pandas as pd
from fastapi import APIRouter, HTTPException
from starlette import status
from transformers import set_seed

from . import TSFM_ALLOW_LOAD_FROM_HF_HUB, TSFM_MODEL_DIR
from .constants import API_VERSION
from .dirutil import resolve_model_path
from .ftpayloads import (
    AsyncCallReturn,
    BaseTuneInput,
    TinyTimeMixerForecastingTuneInput,
)
from .ioutils import to_pandas
from .tuning_handler import TuningHandler


LOGGER = logging.getLogger(__file__)


class FinetuningRuntime:
    def add_routes(self, app):
        self.router = APIRouter(prefix=f"/{API_VERSION}/finetuning", tags=["finetuning"])
        self.router.add_api_route(
            "/tinytimemixer/forecasting",
            self.finetuning,
            methods=["POST"],
            response_model=AsyncCallReturn,
        )
        app.include_router(self.router)

    def finetuning(self, input: TinyTimeMixerForecastingTuneInput, tuned_model_name: str, output_dir: Path):
        answer, ex = self._finetuning_common(input, tuned_model_name=tuned_model_name, tmp_dir=output_dir)

        if ex is not None:
            import traceback

            traceback.print_exception(ex)
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=repr(ex))

        return answer

    def _finetuning_common(
        self, input_payload: BaseTuneInput, tuned_model_name: str, tmp_dir: Path
    ) -> Tuple[Path, Union[Exception, None]]:
        LOGGER.info("in _forecasting_tuning_workflow")

        # set seed, must be done before model load
        if input_payload.parameters.random_seed:
            set_seed(input_payload.parameters.random_seed)

        model_path = resolve_model_path(TSFM_MODEL_DIR, input_payload.model_id)

        if not model_path:
            LOGGER.info(f"Could not find model at path: {model_path}")
            if TSFM_ALLOW_LOAD_FROM_HF_HUB:
                model_path = input_payload.model_id
                LOGGER.info(f"Attempting to use HuggingFace Hub: {model_path}")
            else:
                return None, RuntimeError(
                    f"Could not load model {input_payload.model_id} from {TSFM_MODEL_DIR}. If trying to load directly from the HuggingFace Hub please ensure that `TSFM_ALLOW_LOAD_FROM_HF_HUB=1`"
                )

        handler, e = TuningHandler.load(model_id=input_payload.model_id, model_path=model_path)
        if e is not None:
            return None, e

        parameters = input_payload.parameters
        schema = input_payload.schema

        data: pd.DataFrame = to_pandas(uri=input_payload.data, **schema.model_dump())

        # validation_data = FinetuningRuntime._validation_data(input)

        _, e = handler.prepare(data=data, schema=schema, parameters=parameters)
        if e is not None:
            return None, e

        output, e = handler.train(
            data=data, schema=schema, parameters=parameters, tuned_model_name=tuned_model_name, tmp_dir=tmp_dir
        )
        if e is not None:
            return None, e

        return output, None



"""Functions for generating model-specific parameter payloads"""

# Standard
import importlib
import inspect
import logging
from types import NoneType
from typing import Optional, Set, Union, get_args

import pydantic

# Third Party
from pydantic import BaseModel, ConfigDict, Field


LOGGER = logging.getLogger(__file__)


def get_pydantic_parameters(
    base_name: str,
    config_class_path: str,
    include_set: Optional[Set[str]] = None,
    exclude_set: Optional[Set[str]] = None,
) -> pydantic.BaseModel:
    """Creates a pydantic parameter model for a given HF config class.

    Only one of `include_set` or `exclude_set` should be set. If None they have no effect.

    Args:
        base_name (str): Base configuration object name used to indentify configuration object.
        config_class_path (str): Complete path to module containing configuration class.
        include_set (Optional[Set[str]], optional): Set of parameters to include. Defaults to None.
        exclude_set (Optional[Set[str]], optional): Set of parameters to exclude. Defaults to None.

    Returns:
        pydantic.BaseModel: Resulting parameter model
    """

    if include_set is not None and exclude_set is not None:
        raise ValueError("Only one of `include_set` or `exclude_set` should be specified.")

    config_class_name = f"{base_name}Config"

    module = importlib.import_module(config_class_path)
    config_class = getattr(module, config_class_name)

    params = _create_pydantic_model_from_hf_config(
        config_class,
        f"{base_name}Parameters",
        include_set=include_set,
        exclude_set=exclude_set,
    )
    return params


def _create_pydantic_model_from_hf_config(
    config_class: type,
    model_name: str,
    include_set: Optional[Set[str]] = None,
    exclude_set: Optional[Set[str]] = None,
) -> pydantic.BaseModel:
    """Helper function to create a pydantic model from a HuggingFace config class.

    Args:
        config_class (type): HF config class.
        model_name (str): Name of the resulting model.

    Returns:
        type[BaseModel]: A pydantic parameter model
    """
    field_definitions = {}

    sig = inspect.signature(config_class.__init__)
    for name, param in sig.parameters.items():
        if (
            name in ["self", "kwargs"]
            or ((include_set is not None) and (name not in include_set))
            or ((exclude_set is not None) and (name in exclude_set))
        ):
            continue

        if NoneType in get_args(param.annotation):
            # the magic happens
            LOGGER.debug(f"Rewriting optional parameter {name}")

            new_args = tuple([arg for arg in get_args(param.annotation) if arg is not NoneType])
            # to do: reset defaults appropriately
            if len(new_args) > 1:
                field_definitions[name] = (Union[new_args], param.default)
            else:
                field_definitions[name] = (new_args[0], param.default)
        else:
            field_definitions[name] = (param.annotation, param.default)

    return pydantic.create_model(model_name, **field_definitions)


include_set = {
    "distribution_output",
    "loss",
    "attention_dropout",
    "dropout",
    "positional_dropout",
    "path_dropout",
    "ff_dropout",
    "pooling_type",
    "head_dropout",
    "output_range",
    "num_parallel_samples",
}

PatchTSTParameters = get_pydantic_parameters(
    base_name="PatchTST",
    config_class_path="transformers.models.patchtst.configuration_patchtst",
    include_set=include_set,
)


include_set = {
    "num_parallel_samples",
    "dropout",
    "loss",
    "head_dropout",
    "distribution_output",
    "prediction_length",
    "output_range",
    "pooling_type",
}

PatchTSMixerParameters: BaseModel = get_pydantic_parameters(
    base_name="PatchTSMixer",
    config_class_path="transformers.models.patchtsmixer.configuration_patchtsmixer",
    include_set=include_set,
)


# include_set = {
#     "dropout",
#     "loss",
#     "head_dropout",
#     "distribution_output",
#     "num_parallel_samples",
#     # decoder parameters
#     "decoder_num_layers",
#     "decoder_d_model_scale",
#     "decoder_adaptive_patching_levels",
#     "decoder_raw_residual",
#     "decoder_mode",
#     "use_decoder",
#     # forecast channel mixing wit exog support
#     "enable_forecast_channel_mixing",
#     "fcm_gated_attn",
#     "fcm_context_length",
#     "fcm_use_mixer",
#     "fcm_mix_layers",
# }

# TinyTimeMixerParameters: BaseModel = get_pydantic_parameters(
#     base_name="TinyTimeMixer",
#     config_class_path="tsfm.models.tinytimemixer.configuration_tinytimemixer",
#     include_set=include_set,
# )


class TinyTimeMixerParameters(BaseModel):
    model_config = ConfigDict(extra="forbid")

    dropout: float = Field(default=0.2, description="Dropout")
    head_dropout: float = Field(default=0.2, description="Head dropout")
    # check
    loss: str = Field(default="mse", description="Loss")
    # check
    distribution_output: str = Field(
        default="student_t", description="Distribution output specification"
    )  # student_t or normal or negative_binomial
    num_parallel_samples: int = Field(
        default=100,
        description="Number of parallel samples when sampling from the output distribution",
    )
    # check
    decoder_num_layers: int = Field(default=8, description="Number of decoder layers")
    decoder_adaptive_patching_levels: int = Field(default=0, description="Number of adaptive patching levels")
    decoder_raw_residual: bool = Field(
        default=False,
        description="Flag to enable merging of raw embedding with encoder embedding for decoder input",
    )
    decoder_mode: str = Field(
        default="mix_channel",
        description="Use `common_channel` for channel-independent modeling and `mix_channel` for channel-mixing modeling",
    )
    use_decoder: bool = Field(default=True, description="Enable use of the decoder")
    enable_forecast_channel_mixing: bool = Field(
        default=True,
        description="Use forecast channel mixing when forecasting, useful when exogenous features are present",
    )
    fcm_gated_attn: bool = Field(default=True, description="Enable gated attention in the forecast channel mixer")
    fcm_context_length: int = Field(default=1, description="The context length for forecast channel mixing")
    fcm_use_mixer: bool = Field(default=True, description="Use the forecast channel mixer")
    fcm_mix_layers: int = Field(default=2, description="Number of mixing layers for the forecast channel mixer")
    fcm_prepend_past: bool = Field(default=True, description="Prepend last context for forecast reconciliation")



# Copyright contributors to the TSFM project
#
import argparse
import os
import random
import sys

import matplotlib.pyplot as plt
import numpy as np
import torch
from gluonts.time_feature.seasonality import get_seasonality
from transformers import Trainer
from transformers.utils import logging


logger = logging.get_logger(__name__)

CUSTOM_SEASONALITIES = {
    "S": 60,  # 3600,  # 1 hour
    "s": 60,  # 3600,  # 1 hour
    "T": 60,  # 1440,  # 1 day
    "min": 60,  # 1440,  # 1 day
    "H": 24,  # 1 day
    "h": 24,  # 1 day
    "D": 1,  # 1 day
    "W": 1,  # 1 week
    "M": 12,
    "ME": 12,
    "B": 5,
    "Q": 4,
    "QE": 4,
}


class CustomMASETrainer(Trainer):
    def __init__(self, *args, freq=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.seasonal_period = get_seasonality(freq=freq, seasonalities=CUSTOM_SEASONALITIES)
        logger.info(f"Finetuning with MASE loss with freq={freq}, seasonality = {self.seasonal_period}")

    def compute_loss(self, model, inputs, return_outputs=False, num_items_in_batch=None):
        """
        Compute the MASE loss using the seasonal denominator based on in-sample values.
        """
        # Extract in-sample values from inputs
        in_sample_values = inputs["past_values"]  # Ensure in-sample values are passed in the batch

        # Compute the denominator (mean absolute seasonal error)
        diffs = torch.abs(
            in_sample_values[:, self.seasonal_period :, :] - in_sample_values[:, : -self.seasonal_period, :]
        )
        scale_factor = torch.mean(diffs)

        # Forward pass
        outputs = model(**inputs)
        preds = outputs.prediction_outputs
        labels = inputs["future_values"]

        # Compute the numerator of MASE
        numerator = torch.mean(torch.abs(preds - labels))

        # Compute MASE
        mase_loss = numerator / scale_factor

        return (mase_loss, outputs) if return_outputs else mase_loss


def plot_forecast(
    test_data_input,
    test_data_label,
    forecast_samples,
    prediction_length,
    ds_name,
    term,
    out_dir,
    probabilistic: bool = False,
    quantile_keys=[],
):
    test_data_input_list = list(test_data_input)
    test_data_label_list = list(test_data_label)
    ds_name = ds_name.replace("/", "__")
    plot_dir = f"{out_dir}/{ds_name}__{term}__{prediction_length}"
    num_plots = 10
    plot_context = int(2 * prediction_length)
    channel = 0

    # Set a more beautiful style
    plt.style.use("seaborn-v0_8-whitegrid")

    # Adjust figure size and subplot spacing
    assert num_plots >= 1
    num_plots = min(num_plots, len(forecast_samples))
    fig, axs = plt.subplots(num_plots, 1, figsize=(10, 2 * num_plots))
    if num_plots == 1:
        axs = [axs]

    indices = np.random.choice(len(forecast_samples), size=num_plots, replace=False)

    for i, index in enumerate(indices):
        sample = test_data_input_list[index]
        label = test_data_label_list[index]
        if len(sample["target"].shape) == 1:
            sample["target"] = sample["target"].copy()
            sample["target"] = sample["target"].reshape(1, -1)
        if label["target"].ndim == 1:
            label["target"] = label["target"].reshape(1, -1)
        feasible_plot_context = min(plot_context, sample["target"].shape[1])
        ts_y_hat = np.arange(feasible_plot_context, feasible_plot_context + prediction_length)
        if probabilistic:
            if forecast_samples.ndim == 4:
                y_hat = forecast_samples[index, :, :, channel]
            else:
                y_hat = forecast_samples[index, :, :]

        else:
            if len(forecast_samples.shape) == 3:
                y_hat = forecast_samples[index, :, channel]
            else:
                y_hat = forecast_samples[index, :]

        ts_y = np.arange(feasible_plot_context + prediction_length)
        y = label["target"][channel, :]
        x = sample["target"][channel, -feasible_plot_context:]
        y = np.concatenate((x, y), axis=0)
        border = feasible_plot_context
        plot_title = f"Example {indices[i]}"

        # Plot predicted values with a dashed line
        if probabilistic:
            for qi in [0, 8, 9]:
                axs[i].plot(
                    ts_y_hat,
                    y_hat[qi, :],
                    label=f"Predicted: q_{quantile_keys[qi]}",
                    linestyle="--",
                    # color="orange",
                    linewidth=2,
                )
        else:
            axs[i].plot(
                ts_y_hat,
                y_hat,
                label="Predicted",
                linestyle="--",
                color="orange",
                linewidth=2,
            )

        # Plot true values with a solid line
        axs[i].plot(ts_y, y, label="True", linestyle="-", color="blue", linewidth=2)

        # Plot horizon border
        if border is not None:
            axs[i].axvline(x=border, color="r", linestyle="-")

        axs[i].set_title(plot_title)
        axs[i].legend()

    # Adjust overall layout
    plt.tight_layout()

    # Save the plot
    if plot_dir is not None:
        plot_filename = f"ch_{str(channel)}.pdf"
        os.makedirs(plot_dir, exist_ok=True)
        plt.savefig(os.path.join(plot_dir, plot_filename))


def set_seed(seed):
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def get_args():
    # ## Important arguments

    # Create the parser
    parser = argparse.ArgumentParser(description="A simple argparse example.")

    # Add arguments
    parser.add_argument(
        "--out_dir",
        "-od",
        type=str,
        help="Out dir prefix name",
        required=False,
        default="ttm",
    )
    parser.add_argument(
        "--model_path",
        "-mp",
        type=str,
        help="Path to TTM model card",
        required=False,
        default="ibm-granite/granite-timeseries-ttm-r2",
    )
    parser.add_argument(
        "--decoder_mode",
        "-dm",
        type=str,
        help="Decoder mode",
        required=False,
        default="common_channel",
    )
    parser.add_argument(
        "--head_dropout",
        "-hdr",
        type=float,
        help="Head dropout",
        required=False,
        default=0.2,
    )
    parser.add_argument(
        "--fewshot_fraction",
        "-ff",
        type=float,
        help="Fewshot fraction",
        required=False,
        default=0.2,
    )
    parser.add_argument(
        "--huber_delta",
        "-de",
        type=float,
        help="Huber delta",
        required=False,
        default=1.0,
    )
    parser.add_argument(
        "--quantile",
        "-q",
        type=float,
        help="Quantile for pinball loss",
        required=False,
        default=0.5,
    )
    parser.add_argument(
        "--upper_bound_fewshot_samples",
        "-ubfs",
        type=int,
        help="Upper bound fewshot samples",
        required=False,
        default=0,
    )
    parser.add_argument(
        "--freeze_backbone",
        "-fb",
        type=int,
        help="Freeze TTM backbone",
        required=False,
        default=0,
    )
    parser.add_argument(
        "--batch_size",
        "-bs",
        type=int,
        help="Batch Size",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--fewshot_location",
        "-fl",
        type=str,
        help="Fewshot location: rand/end/start",
        required=False,
        default="rand",
    )
    parser.add_argument(
        "--loss",
        "-l",
        type=str,
        help="Finetune loss",
        required=False,
        default="mae",
    )
    parser.add_argument(
        "--learning_rate",
        "-lr",
        type=float,
        help="Learning rate",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--force_short_context",
        "-fsc",
        type=int,
        help="Force short context for short-term series",
        required=False,
        default=0,
    )
    parser.add_argument(
        "--use_exogs",
        "-ue",
        type=int,
        help="Use any available past exogeneous series",
        required=False,
        default=0,
    )
    parser.add_argument(
        "--use_mask",
        "-um",
        type=int,
        help="Use past_observed_mask in TTM",
        required=False,
        default=0,
    )
    parser.add_argument(
        "--min_context_mult",
        "-mcm",
        type=int,
        help="Minimum context (multiple of horizon) needed for forcing short context",
        required=False,
        default=4,
    )
    parser.add_argument(
        "--num_epochs",
        "-ne",
        type=int,
        help="Number of finetune epochs",
        required=False,
        default=20,
    )
    parser.add_argument(
        "--insample_forecast",
        "-if",
        type=int,
        help="insample prediction to generate quantile forcasts",
        required=False,
        default=1,
    )
    parser.add_argument(
        "--insample_use_train",
        "-iut",
        type=int,
        help="Use training data to calculate insample error statistics",
        required=False,
        default=1,
    )
    parser.add_argument(
        "--use_valid_from_train",
        "-uvft",
        type=int,
        help="Use validation from train",
        required=False,
        default=1,
    )
    parser.add_argument(
        "--mem_per_proc",
        "-mpp",
        type=int,
        help="Memory per ray job (in GB)",
        required=False,
        default=80,
    )
    parser.add_argument(
        "--automate_fewshot_fraction",
        "-aff",
        type=int,
        help="Automatically decide fewshot fraction",
        required=False,
        default=1,
    )
    parser.add_argument(
        "--automate_fewshot_fraction_threshold",
        "-afft",
        type=int,
        help="Threshold on number of samples to automatically decide fewshot fraction",
        required=False,
        default=200,
    )
    parser.add_argument(
        "--short_datasets",
        "-sd",
        type=str,
        help="String of short dataset names separated by spaces",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--med_long_datasets",
        "-mld",
        type=str,
        help="String of medium/long dataset names separated by spaces",
        required=False,
        default=None,
    )

    if "ipykernel" in sys.modules:  # Check if running in a Jupyter environment
        # For Jupyter, provide default arguments
        args = parser.parse_args("")
    else:
        args = parser.parse_args()  # Normal command-line parsing

    args.upper_bound_fewshot_samples = bool(args.upper_bound_fewshot_samples)
    args.freeze_backbone = bool(args.freeze_backbone)
    args.force_short_context = bool(args.force_short_context)
    args.use_exogs = bool(args.use_exogs)
    args.use_mask = bool(args.use_mask)
    args.insample_forecast = bool(args.insample_forecast)
    args.insample_use_train = bool(args.insample_use_train)
    args.use_valid_from_train = bool(args.use_valid_from_train)
    args.automate_fewshot_fraction = bool(args.automate_fewshot_fraction)

    return args



"""Utilities for routes"""

# Standard
import base64
import datetime
import io
import logging
import os
import random
import re
from io import BytesIO
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
from transformers import AutoConfig

# Third Party
# First Party
from tsfm_public import TinyTimeMixerConfig


AutoConfig.register("tinytimemixer", TinyTimeMixerConfig)

LOGGER = logging.getLogger(__file__)


def bytes_to_b64(data: bytes) -> str:
    return base64.b64encode(io.BytesIO(data).getbuffer()).decode("utf-8")


def get_bucket_name(s3_uri):
    return re.search("://(.+?)/", s3_uri).group(1)


def get_object_name(s3_uri):
    bn = get_bucket_name(s3_uri=s3_uri)
    return s3_uri[s3_uri.find(bn) + len(bn) + 1 :]


def _to_pandas(
    iscsv: bool, isfeather: bool, buf: Union[str, BytesIO, None] = None, timestamp_column: Union[str, None] = None
):
    if iscsv:
        if os.path.isdir(buf):
            import pyarrow.dataset as ds

            # Path to directory containing multiple CSV files
            dataset = ds.dataset(buf, format="csv")

            # Convert to a single (virtual) PyArrow Table
            return dataset.to_table().to_pandas()

        else:
            return pd.read_csv(buf, parse_dates=[timestamp_column] if timestamp_column else False)
    if isfeather:
        return pd.read_feather(buf)

    raise Exception("should not be here!")


def _split_text_file(
    source: str, target_dir: Union[str, Path], parts: int = 3, has_header: bool = True, shuffle: bool = True
):
    """Utility for spliting a text file into multiple parts. This is mostly for facilitating data prep
    for test cases involving multiple file input.

    Args:
        source (str): Source file
        target_dir (Union[str, Path]): where split files should be placed
        parts (int, optional): How many near equal parts should be created. Defaults to 3.
        has_header (bool, optional): If source has a header row that should be duplicated in each part. Defaults to True.
        shuffle (bool, optional): Whether to randomly shuffle rows (useful to test certain workflows that might assume ordered data). Defaults to True.
    """
    # split our data into three parts
    lines = open(source).readlines()
    header = lines[0] if has_header else None
    lines = lines[1 if header else 0 :]
    # shuffle rows
    if shuffle:
        random.shuffle(lines)
    splitindex = len(lines) // parts
    remainder = len(lines) % parts
    if not os.path.exists(target_dir):
        os.mkdir(target_dir)
    extension = source[-4:]
    filenames = [f"{os.path.basename(source).replace(extension, '')}.part_{x}{extension}" for x in range(parts)]
    filenames = [(Path(target_dir) / fn).as_posix() for fn in filenames]

    for idx, filename in enumerate(filenames):
        with open(filename, "w") as acsvfile:
            start = idx * splitindex
            end = start + splitindex if idx < parts - 1 else start + splitindex + remainder
            if header:
                acsvfile.write(header)
            acsvfile.writelines(lines[start:end])
    return filenames


def _iscsv(uri: str):
    # we got a directory?
    # we require that it contains **only** files ending with CSV
    #                          0123456
    if uri.upper().startswith("FILE://"):
        if os.path.isdir(uri[7:]):
            contents: List[Tuple] = list(os.walk(uri[7:]))
            # we do not allow nested directories
            if len(contents) > 1:
                raise ValueError(f"{uri} must not have subdirectories.")
            files: List[str] = contents[0][2]  # type: ignore
            if not all(x.upper().endswith(".CSV") for x in files):
                raise ValueError(f"All files in {uri} must have the 'cvs' extension.")
            return True
        else:  # uri is NOT a directory
            if any([uri.upper().endswith(".CSV"), uri.upper().endswith(".CSV.GZ")]):
                return True

    # LOGGER.info("returning False")
    return False


def _isfeather(uri):
    return uri.upper().endswith(".FEATHER")


def _readcsv(buf: BytesIO, timestamp_column):
    try:
        return pd.read_csv(buf, parse_dates=[timestamp_column] if timestamp_column else False)
    except Exception as _:
        buf.seek(0)
        return None


def _readfeather(buf: BytesIO, timestamp_column=None):
    try:
        return pd.read_feather(buf)
    except Exception as _:
        buf.seek(0)
        return None


def to_pandas(uri: str, **kwargs) -> pd.DataFrame:
    """Will return a pandas dataframe for a given uri.

    Args:
        uri (str): A uri which describes the location of the data.
        kwargs: use this to pass arguments from your service payload
        that are used in this method


    Returns:
        pd.DataFrame: a pandas DataFrame object
    """
    received_bytes = not isinstance(uri, str)
    # some guardrails (dependeing on deployment we'll want to paramterize these)
    # at present all we're allowing is a local (file://) refernece to csv or feather
    #                               012345
    if uri[0:6].upper().startswith("S3A://"):
        raise NotImplementedError("reading s3 input is currently forbidden.")
    if received_bytes:
        raise NotImplementedError("reading byte input is currently forbidden.")
    if not received_bytes and uri[0:8].upper().startswith("HTTPS://"):
        raise NotImplementedError("reading from https sources is forbidden.")
    if not received_bytes and uri[0:8].upper().startswith("HTTP://"):
        raise NotImplementedError("reading from http sources is forbidden.")

    timestamp_column = kwargs.get("timestamp_column", None)
    # we're reading from file                              0123456
    if not received_bytes and uri[0:7].upper().startswith("FILE://"):
        if not os.path.exists(uri[7:]):
            raise ValueError(f"{uri[7:]} does not exist.")
        return _to_pandas(iscsv=_iscsv(uri), isfeather=_isfeather(uri), buf=uri[7:], timestamp_column=timestamp_column)
    if os.path.exists(uri):
        raise NotImplementedError(f"{uri} is not a proper URI. Absolute or relative path specifies are not allowed.")

    # attempt to read a binary blob in different formats
    bio = BytesIO(base64.b64decode(uri)) if not received_bytes else BytesIO(uri)
    answer = _readfeather(bio)
    if answer is None:
        answer = _readcsv(bio, timestamp_column=timestamp_column)
    if answer is None:
        raise ValueError("Unable to read data input. Check that it exists and is in the correct format.")
    # we're enforcing timestamp column dtypes that you can
    # perform arithematic operations on
    if timestamp_column is not None and not isinstance(
        answer.dtypes[timestamp_column],
        (
            np.dtypes.DateTime64DType,
            np.dtypes.Float64DType,
            np.dtypes.Int64DType,
            np.dtypes.Int32DType,
            np.dtypes.Int16DType,
            np.dtypes.Float32DType,
            np.dtypes.Float16DType,
            datetime.datetime,
            pd.Timestamp,
            pd.DatetimeTZDtype,
            int,
            float,
        ),
    ):
        raise ValueError(
            f"""column '{timestamp_column}' can not be parsed to a datetime or other numeric type.
            pandas assigns it to type {answer.dtypes[timestamp_column]}."""
        )

    return answer


def getredis():
    """Obtain a redis client. You should have the following environ variables
    defined:

    REDIS_IP (defaults to localhost)
    REDIS_PORT (defaults to 6379)
    REDIS_PASSWORD (must be defined)

    Returns:
        StrictRedis: A StrictRedis implementation.
    """
    # Third Party
    from redis import StrictRedis

    redis_host = os.getenv("REDIS_IP", default="localhost")
    # Note that port 6379 can also be grabbed by occupied by ray's gcs services
    redis_port = int(os.getenv("REDIS_PORT", "6379"))
    redis_password = os.getenv("REDIS_PASSWORD")

    # Checkout here for decode_responses https://redis-py-cluster.readthedocs.io/en/stable/
    # return Redis()
    return StrictRedis(
        host=redis_host,
        port=redis_port,
        password=redis_password,
        # decode_responses=True,
    )



"""Miscellaneous utilities for huggingface transformers"""

import importlib

from transformers import AutoConfig


def register_config(model_type: str, model_config_name: str, module_path: str) -> None:
    """Register a configuration for a particular model architecture

    Args:
        model_type (Optional[str], optional): The type of the model, from the model implementation. Defaults to None.
        model_config_name (Optional[str], optional): The name of configuration class for the model. Defaults to None.
        module_path (Optional[str], optional): Python module path that can be used to load the
            config/model. Defaults to None.

    Raises:
        RuntimeError: Raised when the module cannot be imported from the provided module path.
    """
    # example
    # model_type: "tinytimemixer"
    # model_config_name: "TinyTimeMixerConfig"
    # module_path: "tsfm"  # place where config should be importable

    try:
        mod = importlib.import_module(module_path)
        conf_class = getattr(mod, model_config_name, None)
    except ModuleNotFoundError as exc:  # modulenot found, key error ?
        raise RuntimeError(f"Could not load {model_config_name} from {module_path}") from exc

    if conf_class is not None:
        AutoConfig.register(model_type, conf_class)
    else:
        raise RuntimeError(f"Could not find config for {model_config_name}")



# Standard

import json
import os

import numpy as np
import pandas as pd

# Third Party
from locust import FastHttpUser, task
from tsfminference.inference_payloads import (
    ForecastingInferenceInput,
    ForecastingMetadataInput,
    ForecastingParameters,
)


def ts_data_base(series_length: int, num_timeseries: int, num_targets: int) -> pd.DataFrame:
    # Generate a date range
    length = series_length
    date_range = pd.date_range(start="2023-10-01", periods=length, freq="h")

    timeseries = []
    for idx in range(num_timeseries):
        data: dict = {"date": date_range, "ID": str(idx)}
        for idx in range(num_targets):
            data[f"VAL{idx}"] = np.random.rand(series_length)
        timeseries.append(pd.DataFrame(data))

    answer = pd.concat(timeseries, ignore_index=True)
    answer["date"] = answer["date"].astype(int)
    return answer


def forecasting_input_base(model_id: str, series_length: int, num_timeseries: int, num_targets: int) -> dict:
    df: pd.DataFrame = ts_data_base(series_length, num_timeseries, num_targets)
    schema: ForecastingMetadataInput = ForecastingMetadataInput(
        timestamp_column="date", id_columns=["ID"], target_columns=[f"VAL{idx}" for idx in range(num_targets)]
    )
    parameters: ForecastingParameters = ForecastingParameters()
    input: ForecastingInferenceInput = ForecastingInferenceInput(
        model_id=model_id,
        schema=schema,
        parameters=parameters,
        data=df.to_dict(orient="list"),  # this should get replaced in each test case anyway,
    )
    return input.model_dump()


class MyUser(FastHttpUser):
    max_retries = 3  # custom retry count

    @task
    def forecast_synchronous(self):
        forecasting_url = self.host + "/inference/forecasting"
        for attempt in range(self.max_retries):
            try:
                if attempt > 0:
                    print(f"retrying on attempt {attempt + 1}")
                response = self.client.post(forecasting_url, json=self.payload, timeout=1200)
                if response.status_code == 200:
                    break
                else:
                    print(f"Attempt {attempt + 1} failed with status: {response.status_code}")
            except Exception as e:
                print(f"Attempt {attempt + 1} raised exception: {e}")

    def on_start(self):
        model_param_map = {
            "ttm-r1": {"context_length": 512, "prediction_length": 96},
            "ttm-1024-96-r1": {"context_length": 1024, "prediction_length": 96},
            "ttm-r2": {"context_length": 512, "prediction_length": 96},
            "ttm-r2-etth-finetuned": {"context_length": 512, "prediction_length": 96},
            "ttm-r2-etth-finetuned-control": {"context_length": 512, "prediction_length": 96},
            "ttm-1024-96-r2": {"context_length": 1024, "prediction_length": 96},
            "ttm-1536-96-r2": {"context_length": 1536, "prediction_length": 96},
            "ibm/test-patchtst": {"context_length": 512, "prediction_length": 96},
            "ibm/test-patchtsmixer": {"context_length": 512, "prediction_length": 96},
        }

        model_id = os.environ.get("MODEL_ID", "ttm-r1")
        self.payload = forecasting_input_base(
            model_id=model_id,
            num_timeseries=int(os.environ.get("NUM_TIMESERIES", "10")),
            series_length=model_param_map[model_id]["context_length"],
            num_targets=int(os.environ.get("NUM_TARGETS", "1")),
        )

    def on_stop(self):
        # metrics_url = self.host.replace("/v1", "") + "/metrics"
        # print(self.client.post(metrics_url, json=self.payload, timeout=None, retries=10).text)
        print(f"payload length was {len(json.dumps(self.payload)) / 1e6}MB")



# Copyright contributors to the TSFM project
#
"""Classification-specific preprocessor for time series data preparation"""

import copy
import itertools
from collections import defaultdict
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from datasets import Dataset

from .time_series_preprocessor import (
    INTERNAL_ID_COLUMN,
    LabelEncoder,
    ScalerType,
    TimeSeriesProcessorBase,
    TimeSeriesTask,
)
from .util import check_nested_lengths, is_nested_dataframe, join_list_without_repeat


NESTED_ID_COLUMN = "__nested_series_id"


class TimeSeriesClassificationPreprocessor(TimeSeriesProcessorBase):
    """A preprocessor for supporting time series modeling tasks"""

    PROCESSOR_NAME = "preprocessor_config.json"

    def __init__(
        self,
        id_columns: List[str] = [],
        timestamp_column: Optional[str] = None,
        input_columns: List[str] = [],
        label_column: str = None,
        static_categorical_columns: List[str] = [],
        context_length: int = 64,
        scaling: bool = False,
        scaler_type: ScalerType = ScalerType.STANDARD.value,
        scaling_id_columns: Optional[List[str]] = None,
        encode_categorical: bool = True,
        time_series_task: str = TimeSeriesTask.CLASSIFICATION.value,
        **kwargs,
    ):
        """Multi-time series aware data preprocessor. Provides functions for scaling data and facilitates downstream
        operations on time series data, including model training and inference.

        Args:
            id_columns (List[str]): List of column names which identify different time series in a multi-time series input. Defaults to [].
            timestamp_column (Optional[str], optional): The name of the column containing the timestamp of the time series. Defaults to None.
            input_columns (List[str], optional): List of column names which identify the channels in the input, these are the
                columns that will be used to determine the classification. Defaults to [].
            static_categorical_columns (List[str], optional): List of column names which identify categorical-valued channels in the input
                which are fixed over time. Defaults to [].
            context_length (int, optional): The length of the input context window. Defaults to 64.
            prediction_length (Optional[int], optional): The length of the prediction window. Defaults to None.
            scaling (bool, optional): If True, data is scaled. Defaults to False.
            scaler_type (ScalerType, optional): The type of scaling to perform. See ScalerType for available scalers. Defaults to ScalerType.STANDARD.value.
            scaling_id_columns (Optional[List[str]], optional): In some cases we need to separate data by a different set of id_columns
                when determining scaling factors. For the purposes of determining scaling, data will be grouped by the provided columns.
                If None, the `id_columns` will be used. If and empty list ([]), the dataset will be treated as a single group for scaling.
                Defaults to None. This should be a subset of the id_columns.
            encode_categorical (bool, optional): If True any categorical columns will be encoded using ordinal encoding. Defaults to True.
            time_series_task (str, optional): Reserved for future use. Defaults to TimeSeriesTask.FORECASTING.value.

        Raises:
            ValueError: Raised if `id_columns` is not a list.
            ValueError: Raised if `timestamp_column` is not a scalar.
        """
        # note base class __init__ method sets all arguments as attributes

        if not isinstance(id_columns, list):
            raise ValueError(f"Invalid argument provided for `id_columns`: {id_columns}")

        if isinstance(timestamp_column, list):
            raise ValueError(f"`timestamp_column` should not be a list, received: {timestamp_column}")

        if label_column is None:
            raise ValueError("`label_column` must be specified")

        if isinstance(label_column, list):
            raise ValueError(f"`label_column` should not be a list, received: {label_column}")

        self.id_columns = id_columns
        self.timestamp_column = timestamp_column
        self.label_column = label_column
        self.input_columns = list(input_columns)

        self.static_categorical_columns = list(static_categorical_columns)

        self.context_length = context_length
        self.scaling = scaling
        self.encode_categorical = encode_categorical
        self.time_series_task = time_series_task

        self.scaler_type = scaler_type

        # check subset
        if scaling_id_columns is not None:
            if not set(scaling_id_columns).issubset(self.id_columns):
                raise ValueError("`scaling_id_columns` must be a subset of `id_columns`")
            self.scaling_id_columns = scaling_id_columns
        else:
            self.scaling_id_columns = copy.copy(id_columns)

        self.scaler_dict = {}
        self.categorical_encoder = None

        kwargs["processor_class"] = self.__class__.__name__

        super().__init__(**kwargs)

    def _get_columns_to_encode(
        self,
    ) -> List[str]:
        """Returns the columns to perform encoding on, based on the options specified during
        preprocessor init.

        Returns:
            List[str]: List of column names
        """
        cols_to_encode = self.static_categorical_columns
        return cols_to_encode

    def _validate_columns(self):
        """Check column specification parameters

        Raises:
            ValueError: Raised when a given column appears in multiple column specifiers.
        """

        counter = defaultdict(int)

        for c in self.input_columns + self.static_categorical_columns:
            counter[c] += 1

        if max(counter.values()) > 1:
            raise ValueError(
                "A column name should appear only once in `input_columns` and `static_categorical_columns`."
            )

    def _check_dataset(self, dataset: Union[Dataset, pd.DataFrame], check_nested: bool = True):
        super()._check_dataset(dataset)

        if check_nested:
            check_nested_lengths(dataset, self.input_columns)

    def _get_real_valued_dynamic_channels(
        self,
    ) -> List[str]:
        """Helper function to return list of the real-valued dynamic channels (columns)"""
        real_valued_dynamic_columns = join_list_without_repeat(
            self.input_columns,
        )
        return real_valued_dynamic_columns

    @property
    def num_input_channels(
        self,
    ) -> int:
        """Return the number of input channels

        Input channels are defined as those channels in:
            input_columns

        Future support for time-varying categorical may be addded later.
        """
        return len(self._get_real_valued_dynamic_channels())

    @property
    def exogenous_channel_indices(self) -> List[int]:
        """Return the indices of the exogenous columns

        Classification does not yet support additional exogenous columns.

        Probably remove this method.
        """
        return []

    @property
    def categorical_vocab_size_list(self) -> List[int]:
        """Return the static_categorical_column vocabulary sizes."""
        if not self.static_categorical_columns or not self.encode_categorical:
            return None

        if not self.categorical_encoder:
            raise RuntimeError(
                "Vocabulary sizes are only available after training the preprocessor. Please run the `train` method first."
            )

        sizes = []
        for feat, cats in zip(self.categorical_encoder.feature_names_in_, self.categorical_encoder.categories_):
            if feat in self.static_categorical_columns:
                sizes.append(len(cats))

        return sizes

    def _train_label_encoder(self, df: pd.DataFrame):
        self.label_encoder = LabelEncoder()
        self.label_encoder.fit(df[self.label_column])

    def _train_scaler(self, df: pd.DataFrame):
        scaler_class = self._get_scaler_class(self.scaler_type)
        columns_to_scale = self.input_columns

        if self.scaling:
            for name, g in self._get_groups(df):
                self.scaler_dict[name] = scaler_class()
                if self._is_nested:
                    # one scaler for this group, but requires wrangling for training
                    unnested = unnest_transform(df, columns=columns_to_scale)
                    self.scaler_dict[name].fit(unnested[columns_to_scale])
                else:
                    self.scaler_dict[name] = scaler_class()
                    # one scaler per group
                    self.scaler_dict[name].fit(g[columns_to_scale])

    def train(
        self,
        dataset: Union[Dataset, pd.DataFrame],
    ) -> "TimeSeriesClassificationPreprocessor":
        """Train data transformation operations

        Currently iterates over groups defined by id_columns to train the scaler, if enabled.
        This could be generalized to arbitrary sequence of operations to apply to each group.

        The fitted scalers and their parameters are saved in scaler_dict

        Returns: self

        """

        self._check_dataset(dataset)
        df = self._standardize_dataframe(dataset)
        self._validate_columns()

        self._is_nested = is_nested_dataframe(df, column=self.input_columns[0])

        # if self.freq is None:
        #     self._estimate_frequency(df)

        self._train_label_encoder(df)

        if self.encode_categorical:
            self._train_categorical_encoder(df)
            df = self._process_encoding(df.copy())

        if self.scaling:
            self._train_scaler(df)

        self._clean_up_dataframe(df)
        return self

    def _process_label_encoding(self, df: pd.DataFrame):
        cols_to_encode = self.label_column
        if not self.label_encoder:
            raise RuntimeError("Attempt to encode label column, but the encoder has not been trained yet.")
        df[cols_to_encode] = self.label_encoder.transform(df[cols_to_encode])
        return df

    def preprocess(
        self,
        dataset: Union[Dataset, pd.DataFrame],
    ) -> Dataset:
        """Main function used to return preprocessed data"""
        # for now we assume data is already fully loaded
        # eventually we need a strategy for dealing with:
        # 1) lists of references to datasets
        # 2) incremental / batch based processing of datasets to minimize memory impact

        self._check_dataset(dataset)
        df = self._standardize_dataframe(dataset)
        df = self._process_label_encoding(df)
        df = self._process_encoding(df)

        if self.scaling:
            cols_to_scale = self.input_columns  # self._get_other_columns_to_scale()

            if self.scaling and len(self.scaler_dict) == 0:
                # trying to get output, but we never trained the scaler
                raise RuntimeError(
                    "Attempt to get scaled output, but scaler has not yet been trained. Please run the `train` method first."
                )

            # note, we might want an option to return a copy of the data rather than modifying in place
            def scale_func(grp, id_columns, nested=False):
                if isinstance(id_columns, list):
                    name = tuple(grp.iloc[0][id_columns].tolist())
                else:
                    name = grp.iloc[0][id_columns]

                if nested:
                    unnested = unnest_transform(df, columns=cols_to_scale)
                    unnested[cols_to_scale] = self.scaler_dict[name].transform(unnested[cols_to_scale])
                    grp[cols_to_scale] = nest_transform(unnested, columns=cols_to_scale)
                else:
                    grp[cols_to_scale] = self.scaler_dict[name].transform(grp[cols_to_scale])
                return grp

            if self.scaling_id_columns is not None and len(self.scaling_id_columns) > 0:
                id_columns = (
                    self.scaling_id_columns if len(self.scaling_id_columns) > 1 else self.scaling_id_columns[0]
                )
            else:
                id_columns = INTERNAL_ID_COLUMN

            df_out = df.groupby(id_columns, group_keys=False)[df.columns].apply(
                scale_func, id_columns=id_columns, nested=self._is_nested
            )
            df = df_out

        self._clean_up_dataframe(df)
        return df

    def inverse_transform_labels(self, dataset: pd.DataFrame, suffix: Optional[str] = None) -> pd.DataFrame:
        """Inverse transform the labels back to their original values.

        Args:
            dataset (pd.DataFrame): The input dataframe.

        Returns:
            pd.DataFrame: Dataframe with original values in the label_column
        """
        self._check_dataset(dataset, check_nested=False)
        df = self._standardize_dataframe(dataset)

        col_to_transform = self.label_column
        if suffix is not None:
            col_to_transform = f"{self.label_column}{suffix}"

        df[col_to_transform] = self.label_encoder.inverse_transform(df[col_to_transform])
        self._clean_up_dataframe(df)
        return df


def unnest_transform(df: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
    """Unnest a dataframe that contains nested series.

    Args:
        df (pd.DataFrame): Original dataframe, should contain entries which are pd.Series.
        columns (List[str]): Columns which shoudl be unnested series.

    Raises:
        ValueError: Raised when the dataframe does not contain and id column referenced by NESTED_ID_COLUMN

    Returns:
        pd.DataFrame: Resulting dataframe, but with nested series entries.
    """

    # create a row_id column
    order_preserved_columns = [c for c in df.columns if c in columns]
    series_lengths = df[columns[0]].apply(len).to_list()
    unnested = [[i] * series_lengths[i] for i in df.index]  # range(len(df))]
    # flatten
    unnested = [np.asarray(list(itertools.chain.from_iterable(unnested)))]

    for c in order_preserved_columns:
        unnested.append(np.concatenate([d.values for d in df[c].to_list()]))

    unnested = pd.DataFrame(
        dict(
            zip(
                [
                    NESTED_ID_COLUMN,
                ]
                + order_preserved_columns,
                unnested,
            )
        ),
    )
    return unnested


def nest_transform(df: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
    """Nest a dataframe by first splitting the data by group and creating series for each group.
    The new dataframe has these series as row entries. Original order of the columns in the dataframe
    are preserved.

    Args:
        df (pd.DataFrame): Original dataframe, must contain an id column resulting from the unnest_transformation.
        columns (List[str]): Columns for which to create nested series.

    Raises:
        ValueError: Raised when the dataframe does not contain and id column referenced by NESTED_ID_COLUMN

    Returns:
        pd.DataFrame: Resulting dataframe, but with nested series entries.
    """

    order_preserved_columns = [c for c in df.columns if c in columns]
    if NESTED_ID_COLUMN not in df.columns:
        raise ValueError(
            f"nest_transformation requires that the dataset have an existing nested id column named {NESTED_ID_COLUMN}"
        )
    groups = df.groupby(NESTED_ID_COLUMN)
    rows = []
    index = []
    for name, g in groups:
        rows.append([pd.Series(g[c].values) for c in order_preserved_columns])
        index.append(name)

    nested = pd.DataFrame(rows, columns=order_preserved_columns, index=index)
    return nested



import argparse
import os
import re

import pandas as pd


parser = argparse.ArgumentParser(description="TTM pretrain arguments.")
# Adding a positional argument
parser.add_argument(
    "--result_dir",
    "-rd",
    type=str,
    required=True,
    help="Directory containing results after running benchmarking script.",
)
args = parser.parse_args()

# Path to the main folder containing subfolders
main_folder_path = args.result_dir

# List to collect dataframes
all_data = []

# Iterate through all items in the main folder
for folder_name in os.listdir(main_folder_path):
    # Check if the folder name matches the pattern fl-XX_
    print(folder_name)
    match = re.search(r"fl-(\d+)_", folder_name)
    c_match = re.search(r"cl-(\d+)_", folder_name)
    print(match, c_match)
    if match:
        # Extract XX from the folder name
        XX = int(match.group(1))

        cl = int(c_match.group(1))

        print("reading", XX, cl)
        folder_path = os.path.join(main_folder_path, folder_name)

        # Check if results_zero_few.csv exists in the folder
        csv_file_path = os.path.join(folder_path, "results_zero_few.csv")
        if os.path.exists(csv_file_path):
            # Load the CSV file
            df = pd.read_csv(csv_file_path)

            # Add a new column 'FL' with value XX
            df["FL"] = XX
            df["CL"] = cl

            # Append the dataframe to the list
            all_data.append(df)

# Concatenate all dataframes into one
final_df = pd.concat(all_data, ignore_index=True)

custom_order = ["etth1", "etth2", "ettm1", "ettm2", "weather", "electricity", "traffic"]


final_df["dataset_sorted"] = pd.Categorical(final_df["dataset"], categories=custom_order, ordered=True)

final_df = final_df.sort_values(by=["CL", "dataset_sorted", "FL"], ascending=[True, True, True])


# Save to a new CSV file or process further
out_file_name = os.path.basename(os.path.normpath(main_folder_path))
final_df[["dataset", "CL", "FL", "zs_mse", "fs5_mse"]].to_csv(f"combined_{out_file_name}.csv", index=False)

final_df = final_df.drop(columns=["Unnamed: 0", "dataset_sorted", "FL"])
cols = final_df.columns
cols_index = [
    "dataset",
    "CL",
]
cols_others = [_ for _ in cols if _ not in cols_index]
cols_ord = cols_index + cols_others
final_df = final_df[cols_ord]
avg_df = final_df.groupby(["dataset", "CL"], as_index=False, sort=False).mean()
avg_df = avg_df.round(decimals=3)
avg_df["CL"] = avg_df["CL"].astype(int)
avg_df.to_csv(
    f"combined_avg_{out_file_name}.csv",
    index=False,
)


print(f"All CSV files have been combined and saved as combined_{out_file_name}.csv")
print(f"Average scores per-dataset are as combined_avg_{out_file_name}.csv")



from typing import TYPE_CHECKING

# rely on isort to merge the imports
from transformers.utils import (
    OptionalDependencyNotAvailable,
    _LazyModule,
    is_torch_available,
)


_import_structure = {
    "configuration_tspulse": [
        "TSPULSE_PRETRAINED_CONFIG_ARCHIVE_MAP",
        "TSPulseConfig",
    ],
}

try:
    if not is_torch_available():
        raise OptionalDependencyNotAvailable()
except OptionalDependencyNotAvailable:
    pass
else:
    _import_structure["modeling_tspulse"] = [
        "TSPULSE_PRETRAINED_MODEL_ARCHIVE_LIST",
        "TSPulseModel",
        "TSPulseForReconstruction",
        "TSPulseForClassification",
        "TSPulsePreTrainedModel",
    ]


if TYPE_CHECKING:
    from .configuration_tspulse import (
        TSPULSE_PRETRAINED_CONFIG_ARCHIVE_MAP,
        TSPulseConfig,
    )

    try:
        if not is_torch_available():
            raise OptionalDependencyNotAvailable()
    except OptionalDependencyNotAvailable:
        pass
    else:
        from .modeling_tspulse import (
            TSPULSE_PRETRAINED_MODEL_ARCHIVE_LIST,
            TSPulseForClassification,
            TSPulseForReconstruction,
            TSPulseModel,
            TSPulsePreTrainedModel,
        )

else:
    import sys

    sys.modules[__name__] = _LazyModule(__name__, globals()["__file__"], _import_structure, module_spec=__spec__)



# Copyright contributors to the TSFM project
#
# This code is based on the test code for PatchTSMixer in the HuggingFace Transformers Library:
# https://github.com/huggingface/transformers/blob/main/tests/models/patchtsmixer/test_modeling_patchtsmixer.py
"""Testing suite for the PyTorch TinyTimeMixer model."""

import itertools

# import torchinfo
import unittest

import numpy as np
import torch
from parameterized import parameterized
from transformers import set_seed

from tsfm_public.models.tinytimemixer import (
    TinyTimeMixerConfig,
    TinyTimeMixerForMaskedPrediction,
    TinyTimeMixerForPrediction,
    TinyTimeMixerModel,
)
from tsfm_public.models.tinytimemixer.modeling_tinytimemixer import (
    TinyTimeMixerForPredictionOutput,
)


set_seed(42)


# # Local
# from ...test_configuration_common import ConfigTester
# from ...test_modeling_common import ModelTesterMixin, floats_tensor, ids_tensor
# from ...test_pipeline_mixin import PipelineTesterMixin


# # from transformers.tests.test_configuration_common import ConfigTester
# from transformers.tests.test_modeling_common import (
#     ModelTesterMixin,
#     floats_tensor,
#     ids_tensor,
# )
# # from transformers.tests.test_pipeline_mixin import PipelineTesterMixin


TOLERANCE = 1e-4


class TinyTimeMixerFunctionalTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Setup method: Called once before test-cases execution"""
        cls.params = {}
        cls.params.update(
            context_length=32,
            patch_length=8,
            num_input_channels=3,
            patch_stride=8,
            d_model=64,
            expansion_factor=2,
            num_layers=3,
            adaptive_patching_levels=0,
            dropout=0.2,
            mode="common_channel",  # common_channel,  mix_channel
            gated_attn=True,
            norm_mlp="LayerNorm",
            head_dropout=0.2,
            prediction_length=64,
            # num_labels=3,
            scaling="std",
            use_positional_encoding=False,
            self_attn=False,
            self_attn_heads=1,
            num_parallel_samples=4,
            decoder_num_layers=1,
            decoder_d_model=32,
            decoder_adaptive_patching_levels=0,
            decoder_raw_residual=False,
            decoder_mode="mix_channel",
            use_decoder=True,
        )

        cls.num_patches = (
            max(cls.params["context_length"], cls.params["patch_length"]) - cls.params["patch_length"]
        ) // cls.params["patch_stride"] + 1

        cls.num_masked_patches = (
            max(
                cls.params["context_length"] + cls.params["prediction_length"],
                cls.params["patch_length"],
            )
            - cls.params["patch_length"]
        ) // cls.params["patch_stride"] + 1

        # batch_size = 32
        batch_size = 2
        cls.batch_size = batch_size
        int(cls.params["prediction_length"] / cls.params["patch_length"])

        cls.data = torch.rand(
            batch_size,
            cls.params["context_length"],
            cls.params["num_input_channels"],
        )

        cls.constant_data = torch.ones(
            batch_size,
            cls.params["context_length"],
            cls.params["num_input_channels"],
        )

        cls.freq_token = torch.ones(batch_size)

        cls.enc_data = torch.rand(
            batch_size,
            cls.params["num_input_channels"],
            cls.num_patches,
            cls.params["patch_length"],
        )

        cls.enc_output = torch.rand(
            batch_size,
            cls.params["num_input_channels"],
            cls.num_patches,
            cls.params["d_model"],
        )

        cls.enc_masked_output = torch.rand(
            batch_size,
            cls.params["num_input_channels"],
            cls.num_masked_patches,
            cls.params["d_model"],
        )

        cls.dec_output = torch.rand(
            batch_size,
            cls.params["num_input_channels"],
            cls.num_patches,
            cls.params["decoder_d_model"],
        )

        cls.dec_masked_output = torch.rand(
            batch_size,
            cls.params["num_input_channels"],
            cls.num_masked_patches,
            cls.params["decoder_d_model"],
        )

        cls.flat_enc_output = torch.rand(
            batch_size,
            cls.num_patches,
            cls.params["d_model"],
        )

        cls.correct_pred_output = torch.rand(
            batch_size,
            cls.params["prediction_length"],
            cls.params["num_input_channels"],
        )
        # cls.correct_regression_output = torch.rand(
        #     batch_size, cls.params["num_targets"]
        # )

        cls.correct_pretrain_output = torch.rand(
            batch_size,
            cls.params["num_input_channels"],
            cls.num_patches,
            cls.params["patch_length"],
        )

        cls.correct_forecast_output = torch.rand(
            batch_size,
            cls.params["prediction_length"],
            cls.params["num_input_channels"],
        )

        cls.correct_sel_forecast_output = torch.rand(batch_size, cls.params["prediction_length"], 2)

    def test_patchmodel(self):
        config = TinyTimeMixerConfig(**self.__class__.params)
        mdl = TinyTimeMixerModel(config)
        output = mdl(self.__class__.data)
        self.assertEqual(output.last_hidden_state.shape, self.__class__.enc_output.shape)
        self.assertEqual(output.patch_input.shape, self.__class__.enc_data.shape)

    # def test_forecast_head(self):
    #     config = TinyTimeMixerConfig(**self.__class__.params)
    #     head = TinyTimeMixerForPredictionHead(
    #         config=config,
    #     )
    #     # output = head(self.__class__.enc_output, raw_data = self.__class__.correct_pretrain_output)
    #     output = head(self.__class__.enc_output)

    #     self.assertEqual(output.shape, self.__class__.correct_forecast_output.shape)

    def generate_mask(
        self,
        shape,
        dtype=None,
        obs_prob: float = 0.8,
    ):
        mask_bool = torch.randint(0, 101, shape) < obs_prob * 100
        if dtype is None:
            return mask_bool.type(torch.int64)

        return mask_bool.type(dtype)

    def check_module(
        self,
        task,
        params=None,
        output_hidden_states=True,
        mask_prediction=False,
        future_observed_mask=None,  # None, int, bool
        past_observed_mask=None,  # None, int, bool
        input_data=None,
    ):
        if input_data is None:
            input_data = self.__class__.data
        config = TinyTimeMixerConfig(**params)
        if task == "forecast":
            if mask_prediction is False:
                mdl = TinyTimeMixerForPrediction(config)
            else:
                mdl = TinyTimeMixerForMaskedPrediction(config)
            if (
                "target_channel_filtered" in params
                and params["target_channel_filtered"]
                and params["prediction_channel_indices"] is not None
            ):
                target_input = self.__class__.correct_sel_forecast_output
            else:
                target_input = self.__class__.correct_forecast_output

            if config.prediction_channel_indices is not None:
                target_output = self.__class__.correct_sel_forecast_output
            else:
                target_output = target_input

            if config.prediction_filter_length is not None:
                target_output = target_output[:, : config.prediction_filter_length, :]

            if "target_pred_length_filtered" in params and params["target_pred_length_filtered"]:
                target_input = target_input[:, : config.prediction_filter_length, :]

            ref_samples = target_output.unsqueeze(1).expand(-1, config.num_parallel_samples, -1, -1)

        else:
            print("invalid task")

        if future_observed_mask == "int":
            future_observed_mask = self.generate_mask(shape=target_input.shape)
        elif future_observed_mask == "bool":
            future_observed_mask = self.generate_mask(shape=target_input.shape, dtype=torch.bool)
        else:
            future_observed_mask = None

        if past_observed_mask == "int":
            past_observed_mask = self.generate_mask(shape=input_data.shape)
        elif past_observed_mask == "bool":
            past_observed_mask = self.generate_mask(shape=input_data.shape, dtype=torch.bool)
        else:
            past_observed_mask = None

        if mask_prediction is False:
            enc_output = self.__class__.enc_output
        else:
            enc_output = self.__class__.enc_masked_output

        if config.use_decoder:
            if mask_prediction is False:
                dec_output = self.__class__.dec_output
            else:
                dec_output = self.__class__.dec_masked_output
        else:
            dec_output = enc_output

        cat_samples = None
        if "categorical_vocab_size_list" in params and params["categorical_vocab_size_list"]:
            b = self.__class__.batch_size
            cat_samples = [torch.randint(0, a, (b, 1)) for a in params["categorical_vocab_size_list"]]
            # for i in cat_samples:
            #     print(i.shape,"jjj")
            cat_samples = torch.stack(cat_samples, dim=1).squeeze()
            # print(cat_samples.shape)
            # print(cat_samples)

        if target_input is None:
            output = mdl(
                input_data,
                output_hidden_states=output_hidden_states,
                static_categorical_values=cat_samples,
                future_observed_mask=future_observed_mask,
                past_observed_mask=past_observed_mask,
            )
        else:
            output = mdl(
                input_data,
                future_values=target_input,
                output_hidden_states=output_hidden_states,
                freq_token=self.__class__.freq_token,
                static_categorical_values=cat_samples,
                future_observed_mask=future_observed_mask,
                past_observed_mask=past_observed_mask,
            )

        if isinstance(output.prediction_outputs, tuple):
            for t in output.prediction_outputs:
                self.assertEqual(t.shape, target_output.shape)
        else:
            self.assertEqual(output.prediction_outputs.shape, target_output.shape)

        # self.assertEqual(output.last_hidden_state.shape, enc_output.shape)

        # if output_hidden_states is True:
        #     self.assertEqual(len(output.hidden_states), params["num_layers"])

        # else:
        #     self.assertEqual(output.hidden_states, None)

        if config.loss is not None:
            self.assertEqual(output.loss.item() < np.inf, True)
        enc_output_shape = list(enc_output.shape)
        dec_output_shape = list(dec_output.shape)
        if config.resolution_prefix_tuning:
            enc_output_shape[-2] += 1
            dec_output_shape[-2] += 1

        if mask_prediction is False:
            self.assertEqual(list(output.backbone_hidden_state.shape), enc_output_shape)
            self.assertEqual(list(output.decoder_hidden_state.shape), dec_output_shape)

        # self.assertEqual(output.backbone_hidden_state.shape, enc_output.shape)
        # self.assertEqual(output.decoder_hidden_state.shape, dec_output.shape)

        if config.loss == "nll" and task in ["forecast", "regression"]:
            samples = mdl.generate(input_data)
            self.assertEqual(samples.sequences.shape, ref_samples.shape)

    @parameterized.expand(
        list(
            itertools.product(
                ["common_channel", "mix_channel"],
                [True, False],
                [True, False, "mean", "std"],
                [True, False],
                [None, [0, 2]],
                ["mse", "mae", "pinball", "huber", None],
                [8, 16],
                [True, False],
                [True, False],
            )
        )
    )
    def test_forecast(
        self,
        mode,
        self_attn,
        scaling,
        gated_attn,
        prediction_channel_indices,
        loss,
        prediction_filter_length,
        target_pred_length_filtered,
        mask_prediction,
    ):
        params = self.__class__.params.copy()
        params.update(
            mode=mode,
            self_attn=self_attn,
            scaling=scaling,
            prediction_channel_indices=prediction_channel_indices,
            gated_attn=gated_attn,
            loss=loss,
            prediction_filter_length=prediction_filter_length,
            target_pred_length_filtered=target_pred_length_filtered,
            target_channel_filtered=False,
        )
        self.check_module(task="forecast", params=params, mask_prediction=mask_prediction)

    def test_var0_mask(self):
        params = self.__class__.params.copy()
        self.check_module(
            task="forecast",
            params=params,
            mask_prediction=False,
            future_observed_mask="bool",
            past_observed_mask="bool",
            input_data=self.__class__.constant_data,
        )

    @parameterized.expand(
        list(
            itertools.product(
                [None, [0, 2]],
                ["mse", "mae", None],
                [8, 16],
                [True, False],
                [True, False],
                [None, "int", "bool"],
                [None, "int", "bool"],
            )
        )
    )
    def test_observed_mask(
        self,
        prediction_channel_indices,
        loss,
        prediction_filter_length,
        target_pred_length_filtered,
        mask_prediction,
        past_observed_mask,
        future_observed_mask,
    ):
        params = self.__class__.params.copy()
        params.update(
            prediction_channel_indices=prediction_channel_indices,
            loss=loss,
            prediction_filter_length=prediction_filter_length,
            target_pred_length_filtered=target_pred_length_filtered,
            target_channel_filtered=False,
        )

        self.check_module(
            task="forecast",
            params=params,
            mask_prediction=mask_prediction,
            future_observed_mask=future_observed_mask,
            past_observed_mask=past_observed_mask,
        )

    @parameterized.expand(
        list(
            itertools.product(
                [None, [0, 2]],
                [True, False],
                [True, False],
                ["common_channel", "mix_channel"],
                [0, 2],
                [-1, 3],
                [True, False],
                [None, [1]],
                [True, False],
                [True, False],
            )
        )
    )
    def test_forecast_decoder(
        self,
        prediction_channel_indices,
        use_decoder,
        decoder_raw_residual,
        decoder_mode,
        adaptive_patching_levels,
        decoder_adaptive_patching_levels,
        resolution_prefix_tuning,
        exogenous_channel_indices,
        enable_forecast_channel_mixing,
        fcm_prepend_past,
    ):
        params = self.__class__.params.copy()
        params.update(
            mode="mix_channel",
            self_attn=False,
            scaling=True,
            prediction_channel_indices=prediction_channel_indices,
            gated_attn=True,
            use_decoder=use_decoder,
            decoder_raw_residual=decoder_raw_residual,
            decoder_mode=decoder_mode,
            adaptive_patching_levels=adaptive_patching_levels,
            decoder_adaptive_patching_levels=decoder_adaptive_patching_levels,
            resolution_prefix_tuning=resolution_prefix_tuning,
            exogenous_channel_indices=exogenous_channel_indices,
            enable_forecast_channel_mixing=enable_forecast_channel_mixing,
            fcm_prepend_past=fcm_prepend_past,
            target_channel_filtered=False,
        )

        self.check_module(task="forecast", params=params)

    @parameterized.expand(
        list(
            itertools.product(
                [None, [0, 2]],
                [None, [1]],
                [None, [5, 3, 8]],
                [0, 2],
                [True, False],
                [True, False],
            )
        )
    )
    def test_forecast_catog(
        self,
        prediction_channel_indices,
        exogenous_channel_indices,
        categorical_vocab_size_list,
        decoder_adaptive_patching_levels,
        enable_forecast_channel_mixing,
        decoder_raw_residual,
    ):
        params = self.__class__.params.copy()
        params.update(
            mode="mix_channel",
            self_attn=False,
            scaling=True,
            prediction_channel_indices=prediction_channel_indices,
            gated_attn=True,
            use_decoder=True,
            decoder_raw_residual=decoder_raw_residual,
            decoder_mode="mix_channel",
            adaptive_patching_levels=2,
            decoder_adaptive_patching_levels=decoder_adaptive_patching_levels,
            resolution_prefix_tuning=True,
            exogenous_channel_indices=exogenous_channel_indices,
            enable_forecast_channel_mixing=enable_forecast_channel_mixing,
            fcm_prepend_past=True,
            categorical_vocab_size_list=categorical_vocab_size_list,
            target_channel_filtered=True,
        )

        self.check_module(task="forecast", params=params)

    def forecast_full_module(
        self,
        params=None,
        output_hidden_states=False,
        return_dict=None,
        mask_prediction=False,
    ):
        config = TinyTimeMixerConfig(**params)

        if mask_prediction is False:
            mdl = TinyTimeMixerForPrediction(config)
        else:
            mdl = TinyTimeMixerForMaskedPrediction(config)

        target_val = self.__class__.correct_forecast_output

        target_input = self.__class__.correct_forecast_output

        if config.prediction_channel_indices is not None:
            target_val = self.__class__.correct_sel_forecast_output

        if mask_prediction:
            enc_output = self.__class__.enc_masked_output
        else:
            enc_output = self.__class__.enc_output

        if config.use_decoder:
            if mask_prediction:
                dec_output = self.__class__.dec_masked_output
            else:
                dec_output = self.__class__.dec_output
        else:
            dec_output = enc_output

        if config.prediction_filter_length is not None:
            target_val = target_val[:, : config.prediction_filter_length, :]

        if "target_pred_length_filtered" in params and params["target_pred_length_filtered"]:
            target_input = target_input[:, : config.prediction_filter_length, :]

        cat_samples = None
        if "categorical_vocab_size_list" in params and params["categorical_vocab_size_list"]:
            b = self.__class__.batch_size
            cat_samples = [torch.randint(0, a, (b, 1)) for a in params["categorical_vocab_size_list"]]

            cat_samples = torch.stack(cat_samples, dim=1).squeeze()

        output = mdl(
            self.__class__.data,
            future_values=target_input,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            freq_token=self.__class__.freq_token,
            static_categorical_values=cat_samples,
        )
        # print(mdl)
        # from torchsummary import summary

        # device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        # mdl = mdl.to(device)
        # summary(mdl, self.__class__.data.shape)

        if isinstance(output, tuple):
            output = TinyTimeMixerForPredictionOutput(*output)

        if config.loss in ["mse", "mae"]:
            self.assertEqual(output.prediction_outputs.shape, target_val.shape)
        enc_output_shape = list(enc_output.shape)
        dec_output_shape = list(dec_output.shape)
        if config.resolution_prefix_tuning:
            enc_output_shape[-2] += 1
            dec_output_shape[-2] += 1

        if mask_prediction is False:
            self.assertEqual(list(output.backbone_hidden_state.shape), enc_output_shape)
            self.assertEqual(list(output.decoder_hidden_state.shape), dec_output_shape)

        # if output_hidden_states is True:
        #     print("ooo", len(output.hidden_states))
        #     self.assertEqual(len(output.hidden_states), params["num_layers"])

        # else:
        #     self.assertEqual(output.hidden_states, None)
        if config.loss is not None:
            self.assertEqual(output.loss.item() < np.inf, True)

        if config.loss == "nll":
            samples = mdl.generate(self.__class__.data)
            ref_samples = target_val.unsqueeze(1).expand(-1, params["num_parallel_samples"], -1, -1)
            self.assertEqual(samples.sequences.shape, ref_samples.shape)

    def test_forecast_full(self):
        self.check_module(task="forecast", params=self.__class__.params, output_hidden_states=True)
        # self.forecast_full_module(self.__class__.params, output_hidden_states = True)

    def test_forecast_full_2(self):
        params = self.__class__.params.copy()
        params.update(
            mode="mix_channel",
            adaptive_patching_levels=3,
            num_layers=2,
            resolution_prefix_tuning=False,
            enable_forecast_channel_mixing=True,
            fcm_use_mixer=True,
            fcm_prepend_past=True,
            exogenous_channel_indices=[0],
            categorical_vocab_size_list=[5, 10, 29],
            prediction_filter_length=63,
            target_pred_length_filtered=False,
            loss="mse",
        )
        self.forecast_full_module(params, output_hidden_states=True)

    def test_forecast_full_2_with_return_dict(self):
        params = self.__class__.params.copy()
        params.update(
            mode="mix_channel",
        )
        self.forecast_full_module(params, output_hidden_states=True, return_dict=False)

    def test_forecast_full_3(self):
        params = self.__class__.params.copy()
        params.update(
            mode="mix_channel",
        )
        self.forecast_full_module(params, output_hidden_states=True)

    def test_forecast_full_5(self):
        params = self.__class__.params.copy()
        params.update(
            self_attn=True,
            use_positional_encoding=True,
            positional_encoding="sincos",
        )
        self.forecast_full_module(params, output_hidden_states=True)

    def test_forecast_full_4(self):
        params = self.__class__.params.copy()
        params.update(
            mode="mix_channel",
            prediction_channel_indices=[0, 2],
        )
        self.forecast_full_module(params)

    # def test_forecast_full_distributional(self):
    #     params = self.__class__.params.copy()
    #     params.update(
    #         mode="mix_channel",
    #         prediction_channel_indices=[0, 2],
    #         loss="nll",
    #         distribution_output="normal",
    #     )

    #     self.forecast_full_module(params)

    # def test_forecast_full_distributional_2(self):
    #     params = self.__class__.params.copy()
    #     params.update(
    #         mode="mix_channel",
    #         prediction_channel_indices=[0, 2],
    #         loss="nll",
    #         # distribution_output = "normal",
    #     )
    #     self.forecast_full_module(params)

    # def test_forecast_full_distributional_3(self):
    #     params = self.__class__.params.copy()
    #     params.update(
    #         mode="mix_channel",
    #         # prediction_channel_indices=[0, 2],
    #         loss="nll",
    #         distribution_output="normal",
    #     )
    #     self.forecast_full_module(params)

    # def test_forecast_full_distributional_4(self):
    #     params = self.__class__.params.copy()
    #     params.update(
    #         mode="mix_channel",
    #         # prediction_channel_indices=[0, 2],
    #         loss="nll",
    #         distribution_output="normal",
    #     )
    #     self.forecast_full_module(params)



# Copyright contributors to the TSFM project
#

from typing import TYPE_CHECKING

# rely on isort to merge the imports
from transformers.utils import (
    OptionalDependencyNotAvailable,
    _LazyModule,
    is_torch_available,
)


_import_structure = {
    "configuration_tinytimemixer": [
        "TINYTIMEMIXER_PRETRAINED_CONFIG_ARCHIVE_MAP",
        "TinyTimeMixerConfig",
    ],
}

try:
    if not is_torch_available():
        raise OptionalDependencyNotAvailable()
except OptionalDependencyNotAvailable:
    pass
else:
    _import_structure["modeling_tinytimemixer"] = [
        "TINYTIMEMIXER_PRETRAINED_MODEL_ARCHIVE_LIST",
        "TinyTimeMixerModel",
        "TinyTimeMixerForPrediction",
        "TinyTimeMixerForMaskedPrediction",
    ]

    _import_structure["utils_tinytimemixer"] = [
        "get_freq_mapping",
        "get_freq_token",
        "get_freq_vocab_size",
    ]


if TYPE_CHECKING:
    from .configuration_tinytimemixer import (
        TINYTIMEMIXER_PRETRAINED_CONFIG_ARCHIVE_MAP,
        TinyTimeMixerConfig,
    )

    try:
        if not is_torch_available():
            raise OptionalDependencyNotAvailable()
    except OptionalDependencyNotAvailable:
        pass
    else:
        from .modeling_tinytimemixer import (
            TINYTIMEMIXER_PRETRAINED_MODEL_ARCHIVE_LIST,
            TinyTimeMixerForMaskedPrediction,
            TinyTimeMixerForPrediction,
            TinyTimeMixerModel,
        )
        from .utils_tinytimemixer import (
            get_freq_mapping,
            get_freq_token,
            get_freq_vocab_size,
        )

else:
    import sys

    sys.modules[__name__] = _LazyModule(__name__, globals()["__file__"], _import_structure, module_spec=__spec__)



# Copyright contributors to the TSFM project
#

"""Testing suite for the PyTorch TSPulse model."""

import itertools
import math
import unittest

import numpy as np
import torch
from parameterized import parameterized
from torch.utils.data import DataLoader, Dataset

from tsfm_public.models.tspulse import (
    TSPulseConfig,
    TSPulseForClassification,
    TSPulseForReconstruction,
)
from tsfm_public.models.tspulse.utils.helpers import (
    PatchMaskingDatasetWrapper,
    get_embeddings,
    patchwise_stitched_reconstruction,
)


TOLERANCE = 1e-4


class TSPulseFunctionalTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Setup method: Called once before test-cases execution"""
        cls.params = {}
        cls.total_embedding_size = 64

        cls.params.update(
            context_length=16,
            patch_length=2,
            mask_block_length=2,
            num_input_channels=3,
            patch_stride=2,
            expansion_factor=2,
            num_layers=2,
            dropout=0.1,
            mode="common_channel",  # common_channel,  mix_channel
            gated_attn=True,
            norm_mlp="LayerNorm",
            head_dropout=0.1,
            scaling="std",
            use_positional_encoding=False,
            self_attn=False,
            self_attn_heads=1,
            num_parallel_samples=4,
            decoder_num_layers=2,
            decoder_mode="mix_channel",
            d_model_layerwise_scale=[1, 0.75],  # 0.5, 0.5],
            num_patches_layerwise_scale=[1, 0.75],  # , 0.5, 0.5],
            decoder_num_patches_layerwise_scale=[0.75, 1],  #  0.75, 1],
            decoder_d_model_layerwise_scale=[0.75, 1],  # 0.75, 1],
            num_channels_layerwise_scale=[1, 1],  # 1, 1],  # [1, 0.75, 0.5, 0.5],
            decoder_num_channels_layerwise_scale=[
                1,
                1,
            ],  # 1, 1],  # [0.5, 0.5, 0.75, 1],
            d_model=8,
            decoder_d_model=8,
            num_targets=5,
            head_aggregation="max_pool",
            fuse_fft=True,
            patch_register_tokens=2,
            channel_register_tokens=None,
            fft_mask_ratio=0.2,
            fft_mask_strategy="random",
            use_learnable_mask_token=True,
            prediction_length=4,
            fft_time_add_forecasting_pt_loss=False,
            channel_mix_init="identity",
            reconstruction_loss_weight=1,
            masked_reconstruction_loss_weight=1,
            register_mixer_layers=1,
            head_gated_attention_activation="softmax",
            gated_attention_activation="softmax",
            head_attention=False,
            head_reduce_channels=None,
            fft_applied_on="scaled_ts",
        )

        cls.num_patches = (
            max(cls.params["context_length"], cls.params["patch_length"]) - cls.params["patch_length"]
        ) // cls.params["patch_stride"] + 1

        # batch_size = 32
        batch_size = 32
        cls.batch_size = batch_size

        cls.data = torch.rand(
            batch_size,
            cls.params["context_length"],
            cls.params["num_input_channels"],
        )

        cls.enc_data = torch.rand(
            batch_size,
            cls.params["num_input_channels"],
            cls.num_patches,
            cls.params["patch_length"],
        )

        cls.enc_output = torch.rand(
            batch_size,
            int(cls.params["num_channels_layerwise_scale"][-1] * cls.params["num_input_channels"]),
            int(cls.params["num_patches_layerwise_scale"][-1] * cls.num_patches),
            int(cls.params["d_model_layerwise_scale"][-1] * cls.params["d_model"]),
        )

        cls.enc_output_raw = torch.rand(
            batch_size,
            cls.params["num_input_channels"],
            cls.num_patches,
            cls.params["d_model"],
        )

        cls.dec_output = torch.rand(
            batch_size,
            int(cls.params["decoder_num_channels_layerwise_scale"][-1] * cls.params["num_input_channels"]),
            int(cls.params["decoder_num_patches_layerwise_scale"][-1] * cls.num_patches),
            int(cls.params["decoder_d_model_layerwise_scale"][-1] * cls.params["d_model"]),
        )

        cls.correct_reconstruction_output = torch.rand(
            batch_size,
            cls.params["context_length"],
            cls.params["num_input_channels"],
        )

        cls.loc = torch.rand(
            batch_size,
            1,
            cls.params["num_input_channels"],
        )

        cls.correct_classification_output = torch.rand(
            batch_size,
            cls.params["num_targets"],
        )

        cls.future_values = torch.rand(
            batch_size,
            cls.params["prediction_length"],
            cls.params["num_input_channels"],
        )

        cls.correct_classification_classes = torch.randint(0, cls.params["num_targets"], (batch_size,))

    def check_module(
        self,
        task,
        params=None,
        output_hidden_states=True,
        compression=True,
    ):
        config = TSPulseConfig(**params)
        if task == "reconstruction":
            mdl = TSPulseForReconstruction(config)
            target_input = self.__class__.correct_reconstruction_output
            target_output = target_input
        elif task == "classification":
            config.loss = "cross_entropy"
            mdl = TSPulseForClassification(config)
            target_output = self.__class__.correct_classification_output
            target_input = self.__class__.correct_classification_classes

        else:
            print("invalid task")

        if compression:
            enc_output = self.__class__.enc_output
            enc_output_shape = list(enc_output.shape)

            dec_output = self.__class__.dec_output
            dec_output_shape = list(dec_output.shape)
        else:
            enc_output = self.__class__.enc_output_raw
            dec_output = enc_output
            enc_output_shape = list(enc_output.shape)
            dec_output_shape = list(dec_output.shape)

        if config.mode == "common_channel" or task in ["classification"]:
            enc_output_shape[1] = config.num_input_channels  # no compression for these cases

        if config.fuse_fft:
            enc_output_shape[2] *= 2
            dec_output_shape[2] *= 2

        if config.patch_register_tokens is not None:
            enc_output_shape[2] += config.patch_register_tokens

            if task == "reconstruction" or (
                task in ["classification"] and config.classification_mode == "full_embedding"
            ):
                dec_output_shape[2] += config.patch_register_tokens

            if task in ["classification"] and config.classification_mode == "short_embedding":
                dec_output_shape[2] = config.patch_register_tokens

            if task in ["classification"] and config.classification_mode == "time_with_short_fft_embedding":
                dec_output_shape[2] = (dec_output_shape[2] // 2) + config.patch_register_tokens

            if task in [
                "classification",
            ] and config.classification_mode in ["fft_embedding", "time_embedding"]:
                dec_output_shape[2] = dec_output_shape[2] // 2

        if config.channel_register_tokens is not None:
            enc_output_shape[1] += config.channel_register_tokens

        if config.channel_virtual_expand_scale > 1:
            enc_output_shape[1] *= config.channel_virtual_expand_scale

        enc_output = torch.rand(tuple(enc_output_shape)).flatten(start_dim=2)
        dec_output = torch.rand(tuple(dec_output_shape)).flatten(start_dim=2)

        cat_samples = None
        if "categorical_vocab_size_list" in params and params["categorical_vocab_size_list"]:
            b = self.__class__.batch_size
            cat_samples = [torch.randint(0, a, (b, 1)) for a in params["categorical_vocab_size_list"]]
            cat_samples = torch.stack(cat_samples, dim=1).squeeze()

        if task == "reconstruction":
            output = mdl(
                self.__class__.data,
                future_values=self.__class__.future_values,
                output_hidden_states=output_hidden_states,
            )

            self.assertEqual(output.reconstruction_outputs.shape, target_output.shape)
            self.assertEqual(output.backbone_hidden_state.shape, enc_output.shape)
            self.assertEqual(output.past_values.shape, target_output.shape)

            if config.fuse_fft:
                if config.fft_weight > 0:
                    self.assertEqual(output.fft_reconstruction_outputs.shape, target_output.shape)
                    self.assertEqual(output.original_past_values_fft.shape, target_output.shape)
                if config.fft_original_signal_loss_weight > 0:
                    self.assertEqual(output.reconstructed_ts_from_fft.shape, target_output.shape)
                if config.fft_mask_ratio is not None and config.fft_mask_ratio > 0:
                    self.assertEqual(output.fft_mask.shape, target_output.shape)
                if config.enable_fft_prob_loss:
                    if config.fft_prob_length is None:
                        fft_prob_length = config.context_length // 2
                    else:
                        fft_prob_length = config.fft_prob_length

                    self.assertEqual(
                        output.fft_softmax_preds.shape,
                        target_output[:, :fft_prob_length, :].shape,
                    )
                    self.assertEqual(
                        output.original_fft_softmax.shape,
                        target_output[:, :fft_prob_length, :].shape,
                    )

            if config.fft_time_add_forecasting_pt_loss:
                self.assertEqual(output.future_values.shape, self.__class__.future_values.shape)
                self.assertEqual(output.forecast_output.shape, self.__class__.future_values.shape)

            if config.mask_ratio is not None and config.mask_ratio > 0:
                self.assertEqual(output.masked_past_values.shape, target_output.shape)
                self.assertEqual(output.mask.shape, target_output.shape)

            self.assertEqual(output.fft_loss.item() < np.inf, True)
            self.assertEqual(output.reconstruction_loss.item() < np.inf, True)
            self.assertEqual(output.forecast_loss.item() < np.inf, True)
            self.assertEqual(output.reconstructed_ts_from_fft_loss.item() < np.inf, True)

            self.assertEqual(output.masked_reconstruction_loss.item() < np.inf, True)

            if config.fuse_fft:
                pass

        elif task in ["classification"]:
            output = mdl(
                self.__class__.data,
                output_hidden_states=output_hidden_states,
                target_values=target_input,
                static_categorical_values=cat_samples,
            )
            self.assertEqual(output.prediction_outputs.shape, target_output.shape)

        self.assertEqual(output.loss.item() < np.inf, True)

        self.assertEqual(output.backbone_hidden_state.shape, enc_output.shape)

        self.assertEqual(output.loc.shape, self.__class__.loc.shape)
        self.assertEqual(output.scale.shape, self.__class__.loc.shape)
        self.assertEqual(output.decoder_hidden_state.shape, dec_output.shape)

    @parameterized.expand(
        list(
            itertools.product(
                ["common_channel", "mix_channel"],
                ["common_channel", "mix_channel"],
                [True, False],
                [2],
                ["dc", "last"],
                ["mask", "full", "mask_and_full"],
                [True, False],
                [True, False],
                [0, 1],
                ["revin", "std"],
                ["scaled_ts", "raw_ts"],
            )
        )
    )
    def test_reconstruction_2(
        self,
        mode,
        decoder_mode,
        fuse_fft,
        patch_register_tokens,
        fft_remove_component,
        loss_apply_mode,
        enable_fft_prob_loss,
        fft_time_add_forecasting_pt_loss,
        fft_original_signal_loss_weight,
        scaling,
        fft_applied_on,
    ):
        params = self.__class__.params.copy()
        params.update(
            mode=mode,
            decoder_mode=decoder_mode,
            fuse_fft=fuse_fft,
            patch_register_tokens=patch_register_tokens,
            fft_remove_component=fft_remove_component,
            loss_apply_mode=loss_apply_mode,
            enable_fft_prob_loss=enable_fft_prob_loss,
            fft_time_add_forecasting_pt_loss=fft_time_add_forecasting_pt_loss,
            fft_original_signal_loss_weight=fft_original_signal_loss_weight,
            scaling=scaling,
            fft_applied_on=fft_applied_on,
        )

        self.check_module(task="reconstruction", params=params)

        params.update(
            num_patches_layerwise_scale=None,
            num_channels_layerwise_scale=None,
            d_model_layerwise_scale=None,
            decoder_num_patches_layerwise_scale=None,
            decoder_num_channels_layerwise_scale=None,
            decoder_d_model_layerwise_scale=None,
        )
        self.check_module(
            task="reconstruction",
            params=params,
            output_hidden_states=False,
            compression=False,
        )

    @parameterized.expand(
        list(
            itertools.product(
                [True, False],
                [0.2, None],
                [0.2, None],
                ["var_hybrid", "block", "random", "hybrid"],
                [
                    True,
                    False,
                ],
                [
                    True,
                    False,
                ],
                [
                    True,
                    False,
                ],
                [
                    1,
                    3,
                ],
                [
                    True,
                    False,
                ],
            )
        )
    )
    def test_reconstruction_3(
        self,
        fuse_fft,
        mask_ratio,
        fft_mask_ratio,
        mask_type,
        use_learnable_mask_token,
        channel_consistent_masking,
        fft_time_consistent_masking,
        channel_virtual_expand_scale,
        batch_aware_masking,
    ):
        params = self.__class__.params.copy()
        params.update(
            fuse_fft=fuse_fft,
            mask_ratio=mask_ratio,
            fft_mask_ratio=fft_mask_ratio,
            mask_type=mask_type,
            use_learnable_mask_token=use_learnable_mask_token,
            channel_consistent_masking=channel_consistent_masking,
            fft_time_consistent_masking=fft_time_consistent_masking,
            channel_virtual_expand_scale=channel_virtual_expand_scale,
            batch_aware_masking=batch_aware_masking,
        )
        # print(params)
        self.check_module(task="reconstruction", params=params)
        self.check_module(task="reconstruction", params=params, output_hidden_states=False)

    @parameterized.expand(
        list(
            itertools.product(
                ["random", "magnitude"],
                ["plain", "log"],
                [2, None],
                [0, 0.2, None],
                [0, 0.2, None],
                [True, False],
                [True, False],
            )
        )
    )
    def test_reconstruction_4(
        self,
        fft_mask_strategy,
        fft_prob_mode,
        fft_prob_length,
        mask_ratio,
        fft_mask_ratio,
        use_learnable_mask_token,
        revin_affine,
    ):
        params = self.__class__.params.copy()
        params.update(
            fft_mask_strategy=fft_mask_strategy,
            fft_prob_mode=fft_prob_mode,
            fft_prob_length=fft_prob_length,
            mask_ratio=mask_ratio,
            fft_mask_ratio=fft_mask_ratio,
            use_learnable_mask_token=use_learnable_mask_token,
            scaling="revin",
            revin_affine=revin_affine,
        )

        self.check_module(task="reconstruction", params=params)

    @parameterized.expand(
        list(
            itertools.product(
                ["common_channel", "mix_channel"],
                [True, False, "mean", "std"],
                [True, False],
                ["common_channel", "mix_channel"],
                [True, False],
                ["zero", "identity"],
                [None, 3],
                [True, False],
                [None, [{"head_dropout": 0.3}, {"head_dropout": 0.1}]],
                [True, False],
            )
        )
    )
    def test_reconstruction_1(
        self,
        mode,
        scaling,
        gated_attn,
        decoder_mode,
        fuse_fft,
        channel_mix_init,
        register_mixer_layers,
        free_channel_flow,
        hydra_class_head,
        hydra_class_attention,
    ):
        params = self.__class__.params.copy()
        params.update(
            mode=mode,
            scaling=scaling,
            gated_attn=gated_attn,
            decoder_mode=decoder_mode,
            fuse_fft=fuse_fft,
            channel_mix_init=channel_mix_init,
            register_mixer_layers=register_mixer_layers,
            free_channel_flow=free_channel_flow,
            hydra_class_head=hydra_class_head,
            hydra_class_attention=hydra_class_attention,
        )

        self.check_module(task="reconstruction", params=params)

    @parameterized.expand(
        list(
            itertools.product(
                ["common_channel", "mix_channel"],
                [True, False],
                [True, False, "mean", "std"],
                [True, False],
                ["common_channel", "mix_channel"],
                ["max_pool", "avg_pool", None],
                [True, False],
                [2],
                [None],
                [
                    "full_embedding",
                    "long_embedding",
                    "short_embedding",
                    "time_embedding",
                    "fft_embedding",
                    "time_with_short_fft_embedding",
                ],
                [1, 3],
                [True, False],
            )
        )
    )
    def test_classification(
        self,
        mode,
        self_attn,
        scaling,
        gated_attn,
        decoder_mode,
        pooling,
        fuse_fft,
        patch_register_tokens,
        channel_register_tokens,
        classification_mode,
        channel_virtual_expand_scale,
        disable_mask_in_classification_eval,
    ):
        params = self.__class__.params.copy()
        params.update(
            mode=mode,
            self_attn=self_attn,
            scaling=scaling,
            gated_attn=gated_attn,
            decoder_mode=decoder_mode,
            head_aggregation=pooling,
            fuse_fft=fuse_fft,
            patch_register_tokens=patch_register_tokens,
            channel_register_tokens=channel_register_tokens,
            classification_mode=classification_mode,
            channel_virtual_expand_scale=channel_virtual_expand_scale,
            disable_mask_in_classification_eval=disable_mask_in_classification_eval,
        )

        self.check_module(task="classification", params=params)

    def test_reconstruction_full_2(self):
        params = self.__class__.params.copy()
        params.update(
            {
                "context_length": 16,
                "patch_length": 2,
                "mask_block_length": 2,
                "num_input_channels": 3,
                "patch_stride": 2,
                "expansion_factor": 2,
                "num_layers": 2,
                "dropout": 0.1,
                "mode": "common_channel",
                "gated_attn": True,
                "norm_mlp": "LayerNorm",
                "head_dropout": 0.1,
                "scaling": "std",
                "use_positional_encoding": False,
                "self_attn": False,
                "self_attn_heads": 1,
                "num_parallel_samples": 4,
                "decoder_num_layers": 2,
                "decoder_mode": "mix_channel",
                "d_model_layerwise_scale": [1, 0.75],
                "num_patches_layerwise_scale": [1, 0.75],
                "decoder_num_patches_layerwise_scale": [0.75, 1],
                "decoder_d_model_layerwise_scale": [0.75, 1],
                "num_channels_layerwise_scale": [1, 1],
                "decoder_num_channels_layerwise_scale": [1, 1],
                "d_model": 8,
                "decoder_d_model": 8,
                "num_targets": 5,
                "head_aggregation": "max_pool",
                "fuse_fft": True,
                "patch_register_tokens": 2,
                "channel_register_tokens": None,
                "fft_mask_ratio": 0.2,
                "fft_mask_strategy": "random",
                "use_learnable_mask_token": True,
                "prediction_length": 4,
                "fft_time_add_forecasting_pt_loss": False,
                "channel_mix_init": "identity",
                "reconstruction_loss_weight": 1,
                "masked_reconstruction_loss_weight": 1,
                "register_mixer_layers": 1,
                "head_gated_attention_activation": "softmax",
                "gated_attention_activation": "softmax",
                "head_attention": False,
                "head_reduce_channels": None,
                "mask_ratio": 0.2,
                "mask_type": "var_hybrid",
                "channel_consistent_masking": True,
                "fft_time_consistent_masking": True,
                "channel_virtual_expand_scale": 3,
                "batch_aware_masking": True,
            }
        )
        self.check_module(
            task="reconstruction",
            params=params,
            output_hidden_states=True,
        )

    def test_classification_2(self):
        params = self.__class__.params.copy()
        params.update(
            mode="mix_channel",
            # num_layers=4,
            # decoder_num_layers=4,
            fuse_fft=True,
            mask_ratio=0.3,
            fft_mask_ratio=0.3,
            channel_register_tokens=None,
            patch_register_tokens=2,
        )

        self.check_module(
            task="classification",
            params=params,
            output_hidden_states=True,
        )

    @parameterized.expand(
        list(
            itertools.product(
                [True, False],
            )
        )
    )
    def test_hybrid_mask(self, channel_consistent_masking):
        params = self.__class__.params.copy()
        params.update(
            mode="mix_channel",
            fuse_fft=True,
            mask_ratio=0.3,
            fft_mask_ratio=0.3,
            channel_register_tokens=None,
            patch_register_tokens=2,
            mask_type="hybrid",
            num_full_patches_for_hybrid_mask=1,
            channel_consistent_masking=channel_consistent_masking,
        )

        mdl = TSPulseForReconstruction(TSPulseConfig(**params))
        x = self.__class__.data
        patch_size = mdl.config.patch_length
        mask_token_values = mdl.backbone.time_masker.mask_token.clone()
        # future_values=self.__class__.future_values,
        masked_tensor, mask = mdl.backbone.time_masker.hybrid_masking_with_token(
            x,
            params["mask_ratio"],
            patch_size=patch_size,
            num_full_patches_to_mask=params["num_full_patches_for_hybrid_mask"],
        )
        self.assertEqual(masked_tensor.shape, x.shape)
        self.assertEqual(mask.shape, x.shape)

        T = x.shape[1]
        B = x.shape[0]
        C = x.shape[2]
        patch_ids_relative = torch.arange(T) % patch_size  # [T]
        patch_pos = patch_ids_relative.view(1, T, 1).expand(B, T, C)  # [B, T, C]

        expected_token = mask_token_values[patch_pos]  # [B, T, C]
        # Check if all masked values are replaced correctly
        self.assertTrue(torch.allclose(masked_tensor[mask], expected_token[mask]))

        total_elements = B * T * C
        masked_elements = mask.sum().item()
        expected = int(total_elements * params["mask_ratio"])
        self.assertTrue(abs(masked_elements - expected) / total_elements < 0.1)  # within 10%

        masked_tensor, mask = mdl.backbone.time_masker.hybrid_masking_with_token(
            x,
            1,
            patch_size=patch_size,
            num_full_patches_to_mask=params["num_full_patches_for_hybrid_mask"],
        )

        self.assertTrue(torch.all(mask))

        masked_tensor, mask = mdl.backbone.time_masker.hybrid_masking_with_token(
            x,
            0,
            patch_size=patch_size,
            num_full_patches_to_mask=params["num_full_patches_for_hybrid_mask"],
        )

        self.assertTrue(torch.equal(masked_tensor, x))
        self.assertEqual(mask.sum().item(), 0)

    @parameterized.expand(
        list(
            itertools.product(
                [True, False],
            )
        )
    )
    def test_variable_hybrid_mask(self, channel_consistent_masking):
        params = self.__class__.params.copy()
        params.update(
            mode="mix_channel",
            fuse_fft=True,
            mask_ratio=0.3,
            fft_mask_ratio=0.3,
            channel_register_tokens=None,
            patch_register_tokens=2,
            mask_type="var_hybrid",
            full_patch_mask_percentage=0.5,
            channel_consistent_masking=channel_consistent_masking,
        )

        mdl = TSPulseForReconstruction(TSPulseConfig(**params))
        x = self.__class__.data
        patch_size = mdl.config.patch_length
        mask_token_values = mdl.backbone.time_masker.mask_token.clone()

        # === Base call
        masked_tensor, mask = mdl.backbone.time_masker.variable_length_hybrid_masking_with_token(
            tensor=x,
            mask_percentage=params["mask_ratio"],
            patch_size=patch_size,
            full_patch_mask_percentage=params["full_patch_mask_percentage"],
        )
        self.assertEqual(masked_tensor.shape, x.shape)
        self.assertEqual(mask.shape, x.shape)

        T = x.shape[1]
        B = x.shape[0]
        C = x.shape[2]
        patch_ids_relative = torch.arange(T) % patch_size  # [T]
        patch_pos = patch_ids_relative.view(1, T, 1).expand(B, T, C)  # [B, T, C]

        expected_token = mask_token_values[patch_pos]  # [B, T, C]
        self.assertTrue(torch.allclose(masked_tensor[mask], expected_token[mask]))

        total_elements = B * T * C

        masked_elements = mask.sum().item()
        actual_ratio = masked_elements / total_elements
        expected_ratio = params["mask_ratio"]

        self.assertTrue(
            actual_ratio <= expected_ratio + 0.1,
            f"Actual mask ratio {actual_ratio:.4f} > expected {expected_ratio}",
        )

    def test_past_observed_mask(self):
        params = self.__class__.params.copy()
        params.update(
            mode="mix_channel",
            fuse_fft=True,
            mask_ratio=0.3,
            fft_mask_ratio=0.3,
            channel_register_tokens=None,
            patch_register_tokens=2,
            mask_type="hybrid",
            num_full_patches_for_hybrid_mask=1,
        )

        mdl = TSPulseForReconstruction(TSPulseConfig(**params))
        x = self.__class__.data
        B, T, C = x.shape
        patch_size = mdl.config.patch_length
        mask_token_values = mdl.backbone.time_masker.mask_token.clone()

        past_observed_mask = torch.ones(B, T, C, dtype=torch.bool)
        past_observed_mask[:, 0:4, :] = False  # Mark patch 0 (t=0..3) as missing

        model_output = mdl(x, past_observed_mask=past_observed_mask)
        masked_tensor = model_output.masked_past_values
        mask = model_output.mask

        patch_ids_relative = torch.arange(T) % patch_size
        patch_pos = patch_ids_relative.view(1, T, 1).expand(B, T, C)
        expected_token = mask_token_values[patch_pos]
        # === Tests ===
        self.assertEqual(masked_tensor.shape, x.shape)
        self.assertEqual(mask.shape, x.shape)

        # Check masked positions were replaced by correct tokens
        self.assertTrue(torch.allclose(masked_tensor[mask], expected_token[mask]))

        # Check unmasked positions remain unchanged
        self.assertTrue(torch.allclose(masked_tensor[~mask], x[~mask]))

    def test_patchwise_stitched_reconstruction(self):
        params = self.__class__.params.copy()

        params.update(
            context_length=512,
            patch_length=16,
            patch_stride=16,
            fft_time_add_forecasting_pt_loss=False,
            num_input_channels=4,
            mask_type="user",
        )

        model = TSPulseForReconstruction(TSPulseConfig(**params))

        # # Load pre-trained model
        # model = TSPulseForReconstruction.from_pretrained(
        #     "./tspulse_model",
        #     fft_time_add_forecasting_pt_loss=False,
        #     num_input_channels=4,
        #     mask_type="user",
        # ).to("cuda")
        # model.eval()

        B, L, C = 2, 512, 4  # 4 channels with 1x, 2x, 3x, 4x base frequency
        base_freq = 1.0

        t = torch.linspace(0, 2 * math.pi, L).unsqueeze(1)  # [L, 1]
        waves = []

        for c in range(C):
            freq = base_freq * (c + 1)
            wave = torch.sin(freq * t)  # [L, 1]
            waves.append(wave)

        # Stack along channel dimension → [L, C], then expand to [B, L, C]
        past_values = torch.cat(waves, dim=1).unsqueeze(0).repeat(B, 1, 1)  # [B, L, C]

        patch_size = params["patch_length"]
        patchwise_stitched_reconstruction(
            model,
            past_values=past_values,
            patch_size=patch_size,
            keys_to_stitch=["reconstruction_outputs", "fft_reconstruction_outputs"],
            keys_to_aggregate=[
                # # "forecast_output",
                # "fft_reconstruction_outputs",
                # "original_past_values_fft",
                # # "future_values",
                # "original_fft_softmax",
                # "fft_softmax_preds",
            ],
            reconstruct_start=0,
            reconstruct_end=100,  # to get reconstruction of first 100 points.
            debug=False,
        )

    def test_patchmaskingdatasetwrapper(self):
        class DummyTimeSeriesDataset(Dataset):
            def __init__(self, num_samples=5, T=512, C=2):
                self.T = T
                self.C = C
                self.data = [
                    {
                        "past_values": torch.arange(i * T * C, (i + 1) * T * C).view(T, C).float(),
                        "label": i,
                    }
                    for i in range(num_samples)
                ]

            def __len__(self):
                return len(self.data)

            def __getitem__(self, idx):
                return self.data[idx]

        window_length = 100
        patch_length = 16
        num_patches = math.ceil(window_length / patch_length)
        total_samples = 5

        for window_position in ["last", "first"]:
            base_dataset = DummyTimeSeriesDataset(num_samples=total_samples, T=512, C=2)
            wrapper = PatchMaskingDatasetWrapper(
                base_dataset,
                window_length=window_length,
                patch_length=patch_length,
                window_position=window_position,
            )
            loader = DataLoader(wrapper, batch_size=1, shuffle=False)
            assert len(wrapper) == total_samples * num_patches

            prev_pv = None
            pv_count = 0
            mask_positions = []

            for i, batch in enumerate(loader):
                past_values = batch["past_values"][0]
                past_observed_mask = batch["past_observed_mask"][0]

                if prev_pv is None:
                    prev_pv = past_values
                    pv_count = 1
                elif torch.equal(prev_pv, past_values):
                    pv_count += 1
                else:
                    # Check that patch indices were covered in order: always LTR
                    expected = list(range(num_patches))
                    assert (
                        mask_positions == expected
                    ), f"Incorrect patch order for {window_position}: got {mask_positions}, expected {expected}"
                    assert pv_count == num_patches, f"Expected {num_patches} reps, got {pv_count}"
                    prev_pv = past_values
                    pv_count = 1
                    mask_positions = []

                T, C = past_values.shape
                assert past_observed_mask.shape == (T, C)

                # Check where mask is False (i.e. masked)
                mask = past_observed_mask == 0
                masked_rows = (mask.any(dim=1)).nonzero(as_tuple=True)[0]
                assert masked_rows.numel() > 0, "Each sample must have some masked patch"

                start = masked_rows[0].item()

                # FIX: Normalize start to window to get correct patch_idx
                window_start = 0 if window_position == "first" else T - window_length
                relative_start = start - window_start
                patch_idx = relative_start // patch_length
                mask_positions.append(patch_idx)

                # Ensure mask is inside selected window
                assert window_start <= start < window_start + window_length

                # Validate expected masked length
                expected_masked_len = min(patch_length, T - start)
                actual_masked_len = (mask[start : start + expected_masked_len].any(dim=1)).sum().item()
                assert (
                    actual_masked_len == expected_masked_len
                ), f"Expected {expected_masked_len} rows masked, got {actual_masked_len}"

            # Final flush for last group
            expected = list(range(num_patches))
            assert (
                mask_positions == expected
            ), f"Incorrect patch order for {window_position}: got {mask_positions}, expected {expected}"
            assert pv_count == num_patches, f"Expected {num_patches} reps at end, got {pv_count}"

    def test_get_embeddings(self):
        params = self.__class__.params.copy()
        params.update(
            context_length=512,
            patch_length=8,
            patch_stride=8,
            d_model=24,
            decoder_d_model=24,
            patch_register_tokens=10,
            num_input_channels=1,
            mask_type="user",
            d_model_layerwise_scale=[1, 1],
            num_patches_layerwise_scale=[1, 1],
        )
        past_values = torch.randn((1, 512, 1))  # [B, T, C]

        # num_patches=64*2 ((512-8)/8 + 1)=64, concat([time, fft]))
        # backbone_d_model=decoder_d_model=24, time_num_patches=fft_num_patches=64
        model = TSPulseForReconstruction(TSPulseConfig(**params))
        component = "backbone"
        assert get_embeddings(model, past_values, component=component, mode="time").shape[2] == 1536  # 64*24
        assert get_embeddings(model, past_values, component=component, mode="fft").shape[2] == 1536
        assert get_embeddings(model, past_values, component=component, mode="register").shape[2] == 240  # 10*24
        component = "decoder"
        assert get_embeddings(model, past_values, component=component, mode="time").shape[2] == 1536
        assert get_embeddings(model, past_values, component=component, mode="fft").shape[2] == 1536
        assert get_embeddings(model, past_values, component=component, mode="register").shape[2] == 240

        params.update(
            d_model_layerwise_scale=[1, 0.75],  # backbone_d_model_layerwise = [24, 18]
        )
        # backbone_d_model=18
        model = TSPulseForReconstruction(TSPulseConfig(**params))
        component = "backbone"
        assert get_embeddings(model, past_values, component=component, mode="time").shape[2] == 1152  # 64*18
        assert get_embeddings(model, past_values, component=component, mode="fft").shape[2] == 1152
        assert get_embeddings(model, past_values, component=component, mode="register").shape[2] == 180  # 10*18
        # decoder_d_model=24, i.e., d_model_layerwise_scale does not affect decoder_d_model
        component = "decoder"
        assert get_embeddings(model, past_values, component=component, mode="time").shape[2] == 1536
        assert get_embeddings(model, past_values, component=component, mode="fft").shape[2] == 1536
        assert get_embeddings(model, past_values, component=component, mode="register").shape[2] == 240

        params.update(
            num_patches_layerwise_scale=[1, 0.75],  # backbone_num_patches_layerwise = [128, 96]
        )
        # backbone_d_model=18, time_num_patches=48 (=96/2),
        model = TSPulseForReconstruction(TSPulseConfig(**params))
        component = "backbone"
        assert get_embeddings(model, past_values, component=component, mode="time").shape[2] == 864  # 48*18
        assert get_embeddings(model, past_values, component=component, mode="fft").shape[2] == 864
        assert get_embeddings(model, past_values, component=component, mode="register").shape[2] == 180

        params.update(
            d_model=8,
            decoder_d_model=8,
            patch_register_tokens=8,
        )
        model = TSPulseForReconstruction(TSPulseConfig(**params))
        component = "decoder"
        assert get_embeddings(model, past_values, component=component, mode="register").shape[2] == 64  # 8*8



from datetime import datetime

import pandas as pd
from tsfminference.dataframe_checks import check
from tsfminference.inference_payloads import ForecastingMetadataInput


def test_check():
    # all good
    df = pd.DataFrame(columns=["date", "id1", "id2", "val"], data=[[datetime.now(), "A", "B", 1.0]])
    mdi = ForecastingMetadataInput(timestamp_column="date", id_columns=["id1", "id2"], target_columns=["val"])
    rc, ex = check(data=df, schema=mdi.model_dump())
    assert rc == 0

    # non-integer type for id
    df = pd.DataFrame(columns=["date", "id1", "id2", "val"], data=[[datetime.now(), "A", 1.0, 1.0]])
    rc, ex = check(data=df, schema=mdi.model_dump())
    assert rc == 1
    assert str(ex).find("id2") >= 0
    # bad column reference
    df = pd.DataFrame(columns=["date", "id1", "id2", "val"], data=[[datetime.now(), "A", "B", 1.0]])
    mdi = ForecastingMetadataInput(timestamp_column="timestamp", id_columns=["id1", "id2"], target_columns=["val"])
    rc, ex = check(data=df, schema=mdi.model_dump())
    assert str(ex).find("timestamp") >= 0
    assert rc == 1
    # bad column reference for id col
    df = pd.DataFrame(columns=["date", "id1", "id2", "val"], data=[[datetime.now(), "A", "B", 1.0]])
    mdi = ForecastingMetadataInput(timestamp_column="date", id_columns=["id1", "foobar"], target_columns=["val"])
    rc, ex = check(data=df, schema=mdi.model_dump())
    assert str(ex).find("foobar") >= 0
    assert rc == 1
    # bad column reference for target col
    df = pd.DataFrame(columns=["date", "id1", "id2", "val"], data=[[datetime.now(), "A", "B", 1.0]])
    mdi = ForecastingMetadataInput(
        timestamp_column="date",
        id_columns=["id1", "id2"],
        target_columns=["foobar"],
    )
    rc, ex = check(data=df, schema=mdi.model_dump())
    assert str(ex).find("foobar") >= 0
    assert rc == 1
    # non-existent column
    df = pd.DataFrame(columns=["date", "id1", "id2", "val"], data=[[datetime.now(), "A", "B", 1.0]])
    mdi = ForecastingMetadataInput(
        timestamp_column="date", id_columns=["id1", "id2"], target_columns=["val"], observable_columns=["foobar"]
    )
    rc, ex = check(data=df, schema=mdi.model_dump())
    assert str(ex).find("foobar") >= 0
    assert rc == 1
    # duplicate specifier - need disjoint sets
    df = pd.DataFrame(columns=["date", "id1", "id2", "val"], data=[[datetime.now(), "A", "B", 1.0]])
    mdi = ForecastingMetadataInput(
        timestamp_column="date", id_columns=["id1", "id2"], target_columns=["val"], observable_columns=["val"]
    )
    rc, ex = check(data=df, schema=mdi.model_dump())
    assert str(ex).find("val") >= 0
    assert rc == 1
    # duplicate specifier - need disjoint sets
    df = pd.DataFrame(columns=["date", "id1", "id2", "val"], data=[[datetime.now(), "A", "B", 1.0]])
    mdi = ForecastingMetadataInput(
        timestamp_column="date", id_columns=["id1", "id2"], target_columns=["val"], observable_columns=["date"]
    )
    rc, ex = check(data=df, schema=mdi.model_dump())
    assert str(ex).find("date") >= 0
    assert rc == 1



# Copyright contributors to the TSFM project
#
"""Test service support utility"""

import tempfile

import pytest
from transformers import (
    PatchTSTConfig,
    PatchTSTForPrediction,
    TimeSeriesTransformerConfig,
    TimeSeriesTransformerForPrediction,
)

from tsfm_public.toolkit.service_util import save_deployment_package
from tsfm_public.toolkit.time_series_preprocessor import TimeSeriesPreprocessor


def test_save_deployment_package():
    c = PatchTSTConfig(prediction_length=77)
    m = PatchTSTForPrediction(c)
    tsp = TimeSeriesPreprocessor(random=42)

    with tempfile.TemporaryDirectory() as d:
        save_deployment_package(d, m, ts_processor=tsp)
        m_new = PatchTSTForPrediction.from_pretrained(d)
        assert m_new.config.prediction_length == m.config.prediction_length
        tsp_new = TimeSeriesPreprocessor.from_pretrained(d)
        assert tsp_new.random == tsp.random

    c = TimeSeriesTransformerConfig(prediction_length=11)
    m = TimeSeriesTransformerForPrediction(c)

    with tempfile.TemporaryDirectory() as d:
        with pytest.raises(Exception):
            save_deployment_package(d, m)



"""
# TTM zero-shot and few-shot benchmarking on multiple datasets
Pre-trained TTM models will be fetched from the HuggingFace TTM Model Repositories as described below.

1. TTM-Granite-R1 pre-trained models can be found here: [TTM-R1 Model Card](https://huggingface.co/ibm-granite/granite-timeseries-ttm-r1)
2. TTM-Granite-R2 pre-trained models can be found here: [TTM-R2 Model Card](https://huggingface.co/ibm-granite/granite-timeseries-ttm-r2)
3. TTM-Research-Use pre-trained models can be found here: [TTM-Research-Use Model Card](https://huggingface.co/ibm-research/ttm-research-r2)

Every model card has a suite of TTM models. Please read the respective model cards for usage instructions.
"""

## Imports
import math
import tempfile
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from torch.optim import AdamW
from torch.optim.lr_scheduler import OneCycleLR
from transformers import EarlyStoppingCallback, Trainer, TrainingArguments, set_seed
from transformers.integrations import INTEGRATION_TO_CALLBACK

from tsfm_public import TrackingCallback, count_parameters, load_dataset
from tsfm_public.models.tinytimemixer.utils import get_ttm_args
from tsfm_public.toolkit.get_model import get_model
from tsfm_public.toolkit.lr_finder import optimal_lr_finder
from tsfm_public.toolkit.visualization import plot_predictions


warnings.filterwarnings("ignore")

# Arguments
args = get_ttm_args()

# Set seed
set_seed(args.random_seed)

## Important arguments
# Specify model parameters
CONTEXT_LENGTH = args.context_length
FORECAST_LENGTH = args.forecast_length
FREEZE_BACKBONE = True

# Other args
EPOCHS = args.num_epochs
NUM_WORKERS = args.num_workers

# Make sure all the datasets in the following `list_datasets` are
# saved in the `DATA_ROOT_PATH` folder. Or, change it accordingly.
# Refer to the load_datasets() function
# in notebooks/hfdemo/tinytimemixer/utils/ttm_utils.py
# to see how it is used.
DATA_ROOT_PATH = args.data_root_path

# This is where results will be saved
OUT_DIR = args.save_dir

MODEL_PATH = args.hf_model_path

print(f"{'*' * 20} Pre-training a TTM for context len = {CONTEXT_LENGTH}, forecast len = {FORECAST_LENGTH} {'*' * 20}")

## List of benchmark datasets (TTM was not pre-trained on any of these)

if args.datasets is None:
    list_datasets = [
        "etth1",
        "etth2",
        "ettm1",
        "ettm2",
        "weather",
        "electricity",
        "traffic",
        # "exchange",
        # "zafnoo",
        # "solar" # please note that, solar is part of TTM pre-training.
        #         # But, adding here to do in-distribution testing.
        #         # solar results should be ignored for TTM for zero-shot ranking.
    ]

else:
    list_datasets = [dataset.strip() for dataset in args.datasets.split(",")]


all_results = {
    "dataset": [],
    "zs_mse": [],
    "fs5_mse": [],
}

# Loop over data
for DATASET in list_datasets:
    try:
        print()
        print("=" * 100)
        print(
            f"Running zero-shot/few-shot for TTM-{CONTEXT_LENGTH} on dataset = {DATASET}, forecast_len = {FORECAST_LENGTH}"
        )

        print(f"Model will be loaded from {MODEL_PATH}")
        SUBDIR = f"{OUT_DIR}/{DATASET}"

        # Set batch size
        if DATASET == "traffic":
            BATCH_SIZE = 8
        elif DATASET == "electricity":
            BATCH_SIZE = 32
        else:
            BATCH_SIZE = 64

        # Data prep: Get dataset
        _, _, dset_test = load_dataset(
            DATASET,
            CONTEXT_LENGTH,
            FORECAST_LENGTH,
            dataset_root_path=DATA_ROOT_PATH,
            use_frequency_token=args.enable_prefix_tuning,
            enable_padding=False,
        )

        #############################################################
        ##### Use the pretrained model in zero-shot forecasting #####
        #############################################################
        # Load model
        zeroshot_model = get_model(
            model_path=MODEL_PATH, context_length=CONTEXT_LENGTH, prediction_length=FORECAST_LENGTH
        )

        # zeroshot_trainer
        zeroshot_trainer = Trainer(
            model=zeroshot_model,
            args=TrainingArguments(
                output_dir=tempfile.mkdtemp(),
                per_device_eval_batch_size=BATCH_SIZE,
                seed=args.random_seed,
            ),
            eval_dataset=dset_test,
        )

        # evaluate = zero-shot performance
        print("+" * 20, "Test MSE zero-shot", "+" * 20)
        zeroshot_output = zeroshot_trainer.evaluate(dset_test)
        print(zeroshot_output)
        print("+" * 60)

        # Plot

        if args.plot:
            plot_predictions(
                model=zeroshot_trainer.model,
                dset=dset_test,
                plot_dir=SUBDIR,
                num_plots=10,
                plot_prefix="test_zeroshot",
                channel=0,
            )
            plt.close()

        # write results
        all_results["dataset"].append(DATASET)
        all_results["zs_mse"].append(zeroshot_output["eval_loss"])

    except Exception as e:
        print(f"Reason for exception: {e}")
        # write dummy results
        all_results["dataset"].append(DATASET)
        all_results["zs_mse"].append(np.nan)

    ################################################################
    ## Use the pretrained model in few-shot 5% and 10% forecasting #
    ################################################################
    try:
        if args.fewshot == 0:
            raise Exception("fewshot is not enabled")
        for fewshot_percent in [5]:
            # Set learning rate
            learning_rate = None  # `None` value indicates that the optimal_lr_finder() will be used

            print("-" * 20, f"Running few-shot {fewshot_percent}%", "-" * 20)
            # Data prep: Get dataset
            dset_train, dset_val, dset_test = load_dataset(
                DATASET,
                CONTEXT_LENGTH,
                FORECAST_LENGTH,
                fewshot_fraction=fewshot_percent / 100,
                dataset_root_path=DATA_ROOT_PATH,
                use_frequency_token=args.enable_prefix_tuning,
                enable_padding=False,
            )

            # change head dropout to 0.7 for ett datasets
            if "ett" in DATASET:
                finetune_forecast_model = get_model(
                    model_path=MODEL_PATH,
                    context_length=CONTEXT_LENGTH,
                    prediction_length=FORECAST_LENGTH,
                    head_dropout=0.7,
                )
            else:
                finetune_forecast_model = get_model(
                    model_path=MODEL_PATH,
                    context_length=CONTEXT_LENGTH,
                    prediction_length=FORECAST_LENGTH,
                )

            if FREEZE_BACKBONE:
                print(
                    "Number of params before freezing backbone",
                    count_parameters(finetune_forecast_model),
                )

                # Freeze the backbone of the model
                for param in finetune_forecast_model.backbone.parameters():
                    param.requires_grad = False

                # Count params
                print(
                    "Number of params after freezing the backbone",
                    count_parameters(finetune_forecast_model),
                )

            if learning_rate is None:
                learning_rate, finetune_forecast_model = optimal_lr_finder(
                    finetune_forecast_model,
                    dset_train,
                    batch_size=BATCH_SIZE,
                    enable_prefix_tuning=args.enable_prefix_tuning,
                )
                print("OPTIMAL SUGGESTED LEARNING RATE =", learning_rate)

            print(f"Using learning rate = {learning_rate}")

            # This is to save space during exhaustive benchmarking, use specific directory if the saved models are needed
            tmp_dir = tempfile.mkdtemp()

            finetune_forecast_args = TrainingArguments(
                output_dir=tmp_dir,
                overwrite_output_dir=True,
                learning_rate=learning_rate,
                num_train_epochs=EPOCHS,
                do_eval=True,
                evaluation_strategy="epoch",
                per_device_train_batch_size=BATCH_SIZE,
                per_device_eval_batch_size=BATCH_SIZE,
                dataloader_num_workers=NUM_WORKERS,
                report_to=None,
                save_strategy="epoch",
                logging_strategy="epoch",
                save_total_limit=1,
                logging_dir=tmp_dir,  # Make sure to specify a logging directory
                load_best_model_at_end=True,  # Load the best model when training ends
                metric_for_best_model="eval_loss",  # Metric to monitor for early stopping
                greater_is_better=False,  # For loss
                seed=args.random_seed,
            )

            # Create the early stopping callback
            early_stopping_callback = EarlyStoppingCallback(
                early_stopping_patience=10,  # Number of epochs with no improvement after which to stop
                early_stopping_threshold=0.0,  # Minimum improvement required to consider as improvement
            )
            tracking_callback = TrackingCallback()

            # Optimizer and scheduler
            optimizer = AdamW(finetune_forecast_model.parameters(), lr=learning_rate)
            scheduler = OneCycleLR(
                optimizer,
                learning_rate,
                epochs=EPOCHS,
                steps_per_epoch=math.ceil(len(dset_train) / (BATCH_SIZE)),
            )

            finetune_forecast_trainer = Trainer(
                model=finetune_forecast_model,
                args=finetune_forecast_args,
                train_dataset=dset_train,
                eval_dataset=dset_val,
                callbacks=[early_stopping_callback, tracking_callback],
                optimizers=(optimizer, scheduler),
            )
            finetune_forecast_trainer.remove_callback(INTEGRATION_TO_CALLBACK["codecarbon"])

            # Fine tune
            finetune_forecast_trainer.train()

            # Evaluation
            print(
                "+" * 20,
                f"Test MSE after few-shot {fewshot_percent}% fine-tuning",
                "+" * 20,
            )
            fewshot_output = finetune_forecast_trainer.evaluate(dset_test)
            print(fewshot_output)
            print("+" * 60)

            if args.plot:
                # Plot
                plot_predictions(
                    model=finetune_forecast_trainer.model,
                    dset=dset_test,
                    plot_dir=SUBDIR,
                    num_plots=10,
                    plot_prefix=f"test_fewshot_{fewshot_percent}",
                    channel=0,
                )
                plt.close()

            # write results
            all_results[f"fs{fewshot_percent}_mse"].append(fewshot_output["eval_loss"])

    except Exception as e:
        print(f"Reason for exception: {e}")
        fewshot_percent = 5
        # write dummy results
        all_results[f"fs{fewshot_percent}_mse"].append(np.nan)

    df_out = pd.DataFrame(all_results).round(3)
    print(df_out[["dataset", "zs_mse", "fs5_mse"]])
    df_out.to_csv(f"{OUT_DIR}/results_zero_few.csv")
    df_out.to_csv(f"{OUT_DIR}/results_zero_few.csv")



# Copyright contributors to the TSFM project
#
"""Payload definitions for tsfminference"""

from typing import Annotated, Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator


# WARNING: DO NOT IMPORT util here or else you'll get a circular dependency

EverythingPatternedString = Annotated[str, Field(min_length=0, max_length=100, pattern=r"^\S.*\S$|^\S$")]


class BaseMetadataInput(BaseModel):
    model_config = ConfigDict(extra="forbid", protected_namespaces=())

    timestamp_column: str = Field(
        description="A valid column in the data that should be treated as the timestamp."
        " Although not absolutely necessary, if using calendar dates "
        " (simple integer time offsets are also allowed),"
        " users should consider using a format such as ISO 8601 that"
        " includes a UTC offset (e.g., '2024-10-18T01:09:21.454746+00:00')."
        " This will avoid potential issues such as duplicate dates appearing"
        " due to daylight savings change overs. There are many date formats"
        " in existence and inferring the correct one can be a challenge"
        " so please do consider adhering to ISO 8601.",
        pattern=r"^\S.*\S$|^\S$",
        min_length=1,
        max_length=100,
        example="date",
    )
    id_columns: List[EverythingPatternedString] = Field(
        description="Columns that define a unique key for time series."
        " This is similar to a compound primary key in a database table.",
        default_factory=list,
        max_length=10,
        example=["ID1", "ID2"],
        min_length=0,
    )
    scaling_id_columns: Optional[List[EverythingPatternedString]] = Field(
        description="Columns that define unique subsets of the time series data"
        " over which separate scaling factors should be applied. These columns"
        " must be a subset of the id_columns.",
        default=None,
        max_length=10,
        example=["ID2"],
        min_length=0,
    )
    freq: Optional[str] = Field(
        description="A frequency indicator for the given timestamp_column."
        " See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#period-aliases"
        " for a description of the allowed values. If not provided, we will attempt to infer it from the data.",
        default=None,
        pattern=r"^\d*\.?\d*(B|D|W|M|Q|Y|h|min|s|ms|us|ns)?$",
        min_length=0,
        max_length=100,
        example="1h",
    )


class ForecastingMetadataInput(BaseMetadataInput):
    target_columns: List[EverythingPatternedString] = Field(
        default_factory=list,
        max_length=500,
        min_length=0,
        example=["TARGET1", "TARGET2"],
        description="An array of column headings which constitute the target variables in the data."
        " These are the data that will be forecasted.",
    )
    observable_columns: List[EverythingPatternedString] = Field(
        default_factory=list,
        max_length=500,
        min_length=0,
        example=["OBS1", "OBS2"],
        description="An optional array of column headings which identify"
        " the observables in the data. Observables are features (commonly called channels in timeseries forecasting problems)"
        " which we have knowledge about in the past and future. For example, weather"
        " conditions such as temperature or precipitation may be known or estimated in the future"
        " but cannot be changed. This field supports specialized uses of timeseries forecasting"
        " that the average user is unlikely to encounter.",
    )
    control_columns: List[EverythingPatternedString] = Field(
        default_factory=list,
        max_length=500,
        min_length=0,
        example=["CNTRL1", "CNTRL2"],
        description="An optional array of column headings which identify the control channels in the input."
        " Control channels are similar to observable channels, except that future values may be controlled."
        " For example, the discount percentage of a particular product is known and controllable in the future."
        " Similar to observable_columns, control_columns is intended for advanced use cases not typical"
        " in most timeseries forecasting problems.",
    )
    conditional_columns: List[EverythingPatternedString] = Field(
        default_factory=list,
        max_length=500,
        min_length=0,
        example=["CONDL1", "CONDL2"],
        description="An optional array of column headings which constitute the conditional variables."
        " The conditional_columns in the data are those known in the past, but not known in the future.",
    )
    static_categorical_columns: List[EverythingPatternedString] = Field(
        default_factory=list,
        max_length=500,
        min_length=0,
        example=["SCV1", "SCV2"],
        description="An optional array of column headings which identify"
        " categorical-valued channels in the input which are fixed over time.",
    )

    categorical_columns: List[EverythingPatternedString] = Field(
        default_factory=list,
        max_length=500,
        min_length=0,
        example=["CV1", "CV2"],
        description="An optional array of column headings which identify"
        " categorical-valued channels in the input which can vary over time.",
    )


class BaseParameters(BaseModel):
    model_config = ConfigDict(extra="allow", protected_namespaces=())

    inference_batch_size: Optional[int] = Field(
        description="The batch size used during inference."
        " When multiple time series are present, the inference will be"
        " conducted in batches. If not specified, the model default batch"
        " size will be used.",
        default=None,
    )

    @field_validator("inference_batch_size")
    @classmethod
    def check_inference_batch_size(cls, v: Optional[int]) -> float:
        if v is not None and v < 1:
            raise ValueError(
                "If specified, `inference_batch_size` must be an integer >=1."
                " When omitted the model default inference_batch_size will be used."
            )
        return v


class ForecastingParameters(BaseParameters):
    prediction_length: Optional[int] = Field(
        description="The prediction length for the forecast."
        " The service will return this many periods beyond the last"
        " timestamp in the inference data payload."
        " If specified, `prediction_length` must be an integer >=1"
        " and no more than the model default prediction length."
        " When omitted the model default prediction_length will be used.",
        default=None,
    )

    @field_validator("prediction_length")
    @classmethod
    def check_prediction_length(cls, v: int) -> float:
        if v is not None and v < 1:
            raise ValueError(
                "If specified, `prediction_length` must be an integer >=1"
                " and no more than the model default prediction length."
                " When omitted the model default prediction_length will be used."
            )
        return v


class BaseInferenceInput(BaseModel):
    model_config = ConfigDict(extra="forbid", protected_namespaces=())

    model_id: str = Field(
        description="A model identifier.",
        pattern=r"^\S.*\S$",
        min_length=1,
        max_length=256,
        example="ibm/tinytimemixer-monash-fl_96",
    )


class ForecastingInferenceInput(BaseInferenceInput):
    schema: ForecastingMetadataInput = Field(
        description="An object of ForecastingMetadataInput that contains the schema metadata of the data input.",
    )

    parameters: ForecastingParameters = Field(
        description="additional parameters affecting behavior of the forecast.", default_factory=dict
    )

    data: Dict[str, List[Any]] = Field(
        description="A payload of data matching the schema provided."
        " Let's suppose you have columnar data that looks"
        "  like this (this happens to be csv but it could also be pandas data, for example):\n"
        """
         date,ID1,ID2,TARGET1,VAL2
         2024-10-18T01:00:21+00:00,I1,J1,1.05,10.0
         2024-10-18T01:00:22+00:00,I1,J1,1.75,10.1
         2024-10-18T01:00:21+00:00,I1,J2,2.01,12.8
         2024-10-18T01:00:22+00:00,I1,J2,2.13,13.6\n"""
        " If these data are for two timeseries (each beginning at"
        " 2024-10-18T01:00:21 and ending at 2024-10-18T01:00:22)"
        " given by the compound primary key comprised of ID1 and ID2"
        " and you wish to create predictions only for 'TARGET1',"
        " then your data and schema payload would like like this:\n"
        """
        {
            "schema": {
                "timestamp_column": "date",
                "id_columns": [
                    "ID1",
                    "ID2"
                ],
                "target_columns": [
                    "TARGET1"
                ]
            },
            "data": {
                "date": [
                    "2024-10-18T01:00:21+00:00",
                    "2024-10-18T01:00:22+00:00",
                    "2024-10-18T01:00:21+00:00",
                    "2024-10-18T01:00:22+00:00"
                ],
                "ID1": [
                    "I1",
                    "J1",
                    "I1",
                    "J1"
                ],
                "ID2": [
                    "I1",
                    "J2",
                    "I1",
                    "J2"
                ],
                "TARGET1": [
                    1.05,
                    1.75,
                    2.01,
                    2.13
                ],
                "VAL2": [
                    10.0,
                    10.1,
                    12.8,
                    13.6
                ]
            }
        }\n"""
        "Note that we make no mention of `VAL2` in the schema which means that it will"
        " effectively be ignored by the model when making forecasting predictions."
        " If no `target_columns` are specified, then all columns except `timestamp_column`"
        " will be considered to be targets for prediction. Pandas users can generate the"
        " `data` portion of this content by calling DataFrame.to_dict(orient='list')."
        " The service makes a few assumptions about your data:"
        " * All time series are of equal length and are uniform in nature (the time difference between"
        " two successive rows is constant);"
        " * The above implies that there are no missing rows of data;"
        " * You can not have any missing cells of data within in a row (no null NaN values either);"
        " * The above constraints mean that you are responsible for performing your own imputation on your"
        " data before passing it to the service.",
        min_length=1,
    )

    future_data: Optional[Dict[str, List[Any]]] = Field(
        description="Exogenous or supporting features that extend into"
        " the forecasting horizon (e.g., a weather forecast or calendar"
        " of special promotions) which are known in advance."
        " `future_data` would be in the same format as `data` except"
        "  that all timestamps would be in the forecast horizon and"
        " it would not include previously specified target columns."
        " Here's an example payload for such data:\n"
        """
        {
            "future_data": {
                "date": [
                    "2024-10-18T01:00:23+00:00",
                    "2024-10-18T01:00:24+00:00",
                    "2024-10-18T01:00:23+00:00",
                    "2024-10-18T01:00:24+00:00"
                ],
                "ID1": [
                    "I1",
                    "J1",
                    "I1",
                    "J1"
                ],
                "ID2": [
                    "I1",
                    "J2",
                    "I1",
                    "J2"
                ],
                "VAL2": [
                    11.0,
                    11.1,
                    13.8,
                    14.6
                ]
            }
        }\n"""
        "Note that we make no mention of `TARGET1` (from the `data` field example) and"
        " that all timestamps are in the _future_ relative to the `data` you provided."
        " Given these `future_data` the model (when supported) will factor in `VAL2` when"
        " making predictions for `TARGET1`.",
        default=None,
    )


class PredictOutput(BaseModel):
    model_id: str = Field(
        description="Model identifier for the model that produced the prediction.",
        default=None,
    )
    created_at: str = Field(
        description="Timestamp indicating when the prediction was created. ISO 8601 format.",
        default=None,
    )
    results: List[Dict[str, List[Any]]] = Field(
        description="List of prediction results.",
        default=None,
    )

    input_data_points: int = Field(description="Count of input data points.", default=None)
    output_data_points: int = Field(description="Count of output data points.", default=None)
